import copy
import os.path
import subprocess

import SimpleITK as sitk
import nibabel
from pydicom._storage_sopclass_uids import SecondaryCaptureImageStorage, RLELossless
from pydicom.uid import generate_uid

from ._dicom import *


# todo: 2023-05-16: 3D格式的 dicom文件有些问题
# todo: 有时间查看dcm2niix源码



def g_slice_location(first_slice_location, z_space, slice_num):
    for i in range(slice_num):
        yield first_slice_location + i * z_space

def g_position(m, slice_num):
    for i in range(slice_num):
        yield m[:3, 3] + i * m[:3, 2]

def cmd_dcm2niix(src, out, format):
    cmd = f'"D:\Program Files\MRIcron\Resources\dcm2niix.exe" -o "{out}" -z y -f "{format}" -b n -w 1 "{src}"'
    c = subprocess.Popen(cmd)
    c.wait()
    return c.returncode

def save_dicom(dicom_path: str, ds: pydicom.Dataset):
    ds_copy = copy.deepcopy(ds)
    ds_copy.compress(RLELossless)
    ds_copy.file_meta.TransferSyntaxUID = RLELossless
    ds_copy.save_as(dicom_path, write_like_original=False)

def write_dicom_series(image_data:np.ndarray, series_model:SeriesModel, tag:str, series_number:int, output_path:str,
                       slope:float=1, intercept: float=0, replace_lambda=None, dtype=None):
    '''
    将3D图像按照每个切片保存为单个DICOM文件
    :param image_data:
    :param series_model: 大部分标签所参考的对象
    :param tag:
    :param output_path:
    :param slope: 原始数值倍数
    :param intercept:
    :param replace_lambda: 需要额外替换的标签
    :return:
    '''
    # 每个切片保存为一个DICOM文件
    if replace_lambda is None:
        replace_lambda = lambda x: x
    os.makedirs(output_path, exist_ok=True)

    dicom_affine = series_model.calc_affine()
    if series_model.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_MOSAIC:
        # 只考虑了 Axial 的mosaic图像
        normal_direction = dicom_affine[2, 2] / abs(dicom_affine[2, 2]) # 切片方向
        for i in range (series_model.SliceNumber):
            position = (dicom_affine[:3, 3] + i * dicom_affine[:3, 2]).tolist()
            slice_location =  series_model.SliceLocation + normal_direction * series_model.SliceThickness * i
            slice_image = image_data[:,:,i]

            dataset = create_dicom_forpydicom(slice_image, position, series_model.ImageOrientationPatient, slice_location,
                                              series_model, series_number, tag, slope, intercept)
            dataset.InstanceNumber = i + 1
            replace_lambda(dataset)
            save_dicom(os.path.join(output_path, 'MR%06d.dcm'%(i+1)), dataset)

    elif series_model.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_2D:
        # slice_location 增加
        for i in range(series_model.SliceNumber):
            slice_image = image_data[:, :, i]
            if series_model.is_second_arrangement_mode:
                position = series_model[i * series_model.Repetition].ImagePositionPatient
                slice_location = series_model[i * series_model.Repetition].SliceLocation
            else:
                position = series_model[i].ImagePositionPatient
                slice_location = series_model[i].SliceLocation

            dataset = create_dicom_forpydicom(slice_image, position, series_model.ImageOrientationPatient, slice_location,
                                              series_model, series_number, tag, slope, intercept)
            dataset.InstanceNumber = i + 1
            replace_lambda(dataset)
            save_dicom(os.path.join(output_path, 'MR%06d.dcm'%(i+1)), dataset)

    elif series_model.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D:
        first_z_position = series_model.ImagePositionPatientList[0][2]
        second_z_position = series_model.ImagePositionPatientList[1][2]
        normal_direction = 1 if first_z_position < second_z_position else -1
        for i in range(series_model.SliceNumber):
            slice_image = image_data[:, :, i]
            slice_location = series_model.SliceLocation + normal_direction * series_model.SliceThickness * i
            dataset = create_dicom_forpydicom(slice_image, series_model.ImagePositionPatientList[i], series_model.ImageOrientationPatient,
                                              slice_location, series_model, series_number, tag, slope, intercept)
            dataset.InstanceNumber = i + 1
            replace_lambda(dataset)
            save_dicom(os.path.join(output_path, 'MR%06d.dcm'%(i+1)), dataset)

    else:
        raise UnidentifiedDicomImageTypeError()

    nifti_affine = dicom_affine.copy()
    # nifti 与 dicom 方向不一样 dicom: LPS -> nifti: RAS
    nifti_affine[:2] *= -1
    # todo：
    # 自己写的转换nifti，需要大量验证；速度最快；其中转换代码从dcm2niix源码中而来
    image_data_swap = (image_data.swapaxes(0, 1) * slope + intercept)
    if dtype is not None:
        image_data_swap = image_data_swap.astype(dtype)
    elif ('asl' not in tag) or 'm0' in tag:
        image_data_swap = image_data_swap.astype(np.int16)
    else:
        image_data_swap = image_data_swap.astype(np.float32)

    filename = os.path.join(os.path.dirname(output_path), tag + '.nii.gz')
    nibabel.Nifti1Image(image_data_swap, nifti_affine).to_filename(filename)

    # todo 利用sitk修复 nifti文件，自己写的nifti头部可能有些错误，会对后续步骤造成影响
    reader = sitk.ImageFileReader()
    reader.SetImageIO("NiftiImageIO")
    reader.SetFileName(filename)
    new_image = reader.Execute()

    writer = sitk.ImageFileWriter()
    writer.SetFileName(filename)
    writer.Execute(new_image)

    # 用的开源dcm2niix，依赖dcm2niix.exe；最完善的转换插件
    # cmd_dcm2niix(output_path, os.path.dirname(output_path), tag)

    # python库，效率较低，但是无需依赖；也存在些问题
    # dicom2nifti.dicom_series_to_nifti(output_path, output_path + '.nii.gz', reorient_nifti=True)

def create_dicom_forpydicom(image:np.ndarray, position, orientation, slice_location, series_model: SeriesModel,
                            series_number: int, tag: str, slope: float=1, intercept: float=0):
    ds = pydicom.Dataset()
    ds.SOPClassUID = SecondaryCaptureImageStorage
    ds.SOPInstanceUID = generate_uid()
    ds.is_implicit_VR = True
    ds.is_little_endian = True
    ds.fix_meta_info()
    ds.file_meta.FileMetaInformationGroupLength = 1

    ds.PatientName = series_model.PatientName
    ds.PatientID = series_model.PatientID
    ds.PatientBirthDate = series_model.PatientBirthDate
    ds.PatientSex = series_model.PatientSex
    ds.PatientAge = series_model.PatientAge
    ds.PatientSize = series_model.PatientSize
    ds.PatientWeight = series_model.PatientWeight
    ds.PatientPosition = series_model.PatientPosition

    ds.StudyInstanceUID = series_model.StudyInstanceUID
    ds.StudyDescription = series_model.StudyDescription
    ds.StudyDate = series_model.StudyDate
    ds.StudyTime = series_model.StudyTime
    ds.StudyID = series_model.StudyID

    ds.SeriesInstanceUID = series_model.SeriesInstanceUID + '.1'
    ds.SeriesDate = series_model.SeriesDate
    ds.SeriesTime = series_model.SeriesTime
    ds.SeriesNumber = series_number

    ds.Manufacturer = series_model.Manufacturer.value
    ds.ManufacturerModelName = series_model.ManufacturerModelName
    ds.Modality = series_model.Modality.value
    ds.InstitutionName = series_model.InstitutionName

    ds.BitsAllocated = 16
    ds.BitsStored = 16
    ds.HighBit = 16
    ds.SliceThickness = series_model.SliceThickness
    ds.PixelSpacing = series_model.PixelSpacing
    ds.PixelRepresentation = 1
    ds.SpacingBetweenSlices = series_model.SpacingBetweenSlices
    ds.PhotometricInterpretation = series_model.PhotometricInterpretation

    ds.ProtocolName = tag
    ds.SeriesDescription = tag
    ds.SamplesPerPixel = 1
    ds.Rows = image.shape[0]
    ds.Columns = image.shape[1]
    ds.RescaleSlope = slope
    ds.RescaleIntercept = intercept


    # t = datetime.now()
    # ds.InstanceCreationDate = t.strftime('%Y%m%d')
    # ds.InstanceCreationTime = t.strftime('%H%M%S')

    ds.ImagePositionPatient = position
    ds.ImageOrientationPatient = orientation
    ds.SliceLocation = slice_location

    # 溢出的数值警告
    if image.max() > 32767:
        logger.warning('图像数值有溢出，> 32767 将使用32767代替')
        image[image > 32767] = 32767
    if image.min() < -32768:
        logger.warning('图像数值有溢出，< -32768 将使用-32768代替')
        image[image < -32768] = -32768

    ds.PixelData = image.astype(np.int16).tobytes()

    return ds

def create_dicom_forsitk(image:np.ndarray, position, orientation, slice_location, ref: pydicom.Dataset):
    image = sitk.GetImageFromArray(image.astype(np.uint16))
    image.SetMetaData('0020|0037', '\\'.join([str(i) for i in orientation]))
    image.SetMetaData('0020|0032', '\\'.join([str(i) for i in position]))
    image.SetMetaData('0020|1041', str(slice_location))


    image.SetMetaData('0020|000E', ref.SeriesInstanceUID)
    image.SetMetaData('0020|000d', ref.StudyInstanceUID)
    uid = generate_uid()
    image.SetMetaData('0002|0003', uid)
    image.SetMetaData('0008|0018', uid)
    image.SetMetaData('0002|0002', SecondaryCaptureImageStorage)
    image.SetMetaData('0008|0016', SecondaryCaptureImageStorage)
    image.SetMetaData('0008|0060', 'MR')
    image.SetMetaData('0008|0030', ref.StudyTime)
    image.SetMetaData('0008|0020', ref.StudyDate)
    # image.SetMetaData('0028|1052', str(-1))
    # image.SetMetaData('0028|1053', str(0.1))
    return image




"""
处理动脉自旋标记图像的矩阵用的工具包
兼容西门子多延迟、通用电气单延迟、飞利浦单延迟数据
可以计算GE7延迟的mCBF、corr-mCBF、CBF1-7、corr-CBF1-7的功能，需要导入eASL：raw和eASL：transit-delay
准确性：mCBF和CBF1-7没问题，corr-mCBF有点散但是拟合没问题，corr-CBF1-7猜的，建议不要用，用GE自己生成的

TODO：
加降噪功能、加头动矫正功能
王老师序列单延迟输出的确认、完善飞利浦的CBF计算、
完善ATT-mapping以及王老师序列生成每一个CBF的对比

版本：3.0
日期：2021-06-07
作者：姚泽山

2.0 -> 3.0: 增加计算通用电气7延迟的mCBF、CBF1~7
1.0 -> 2.0: 增加计算飞利浦和通用电气的CBF数据，以及Masking参数；现在需要输入核磁厂家信息
"""

import os
import numpy as np
import pydicom as pd
from scipy.optimize import curve_fit


def get_balloon_model_parameters():
    pul_dur = 1.5  # pulse duration
    blc = 0.9  # Blood Tissue Coefficient, lambda
    tag = 0.73  # Tagging Efficiency
    r1a = 0.61
    return pul_dur, blc, tag, r1a


def func(x, att, cbf):
    pul_dur, blc, tag, r1a = get_balloon_model_parameters()
    y = cbf * tag / 3000 / blc / r1a * np.piecewise(x, [x < att, x >= att],
                                                    [lambda x: np.exp(-att * r1a) - np.exp(-r1a * (pul_dur + x)),
                                                     lambda x: np.exp(-x * r1a) - np.exp(-r1a * (pul_dur + x))])
    return y


def binary_search(arr, x):
    n = len(arr)
    if x > arr[-1]:
        return n
    left, right = 0, n - 1
    while left <= right:
        mid = (left + right) // 2
        if x <= arr[mid]:
            right = mid - 1
        else:
            left = mid + 1
    return left


class AnImageASLUtility:

    def __init__(self, vendor=[], series_type=[]):
        self.vendor = vendor
        self.series_type = series_type
        self.raw_data = []
        self.measure_list = []
        self.duration_list = []
        self.delay_list = []
        self.threshold = []

        self.img_m0 = []
        self.mask = []

        self.img_plds = []
        self.img_wd = []
        self.img_att = []
        self.img_cbfs = []
        self.img_diff_cbfs = []
        self.img_mcbf = []
        self.img_acbv = []

        self.number_of_averages = []
        self.magnetic_field_strength = []
        self.labeling_duration = []

        # TODO 额外加上的系数；只是不知为什么跟小彭输出有差异的部分，我猜是平滑导致的
        self.extra_pld_factor = 0.92
        self.extra_wds_factor = 0.90
        self.extra_att_factor = 0.90
        self.extra_cbf_factor = 0.88
        self.extra_cbv_factor = 1.80

    def set_vendor(self, vendor):
        self.vendor = vendor

    def set_series_type(self, series_type):
        self.series_type = series_type

    def set_scan_parameter(self, pydicom_image):
        img = pydicom_image
        self.number_of_averages = img[0x0018, 0x0083].value if [0x0018, 0x0083] in img else 1
        self.magnetic_field_strength = img[0x0018, 0x0087].value if [0x0018, 0x0087] in img else 1
        try:
            self.labeling_duration = img[0x0043, 0x10A5].value / 1000
        except:
            self.labeling_duration = 1.5

    def set_duration_list(self, input_list):
        self.duration_list = input_list

    def get_duration_list(self):
        return self.duration_list

    def set_delay_list(self, *argv):
        if len(argv) == 1:
            self.delay_list = argv[0]

        elif len(argv) == 3:
            init_delay = argv[0]
            delay_num = argv[1]
            delay_interval = argv[2]
            delays = []
            for delay_idx in range(delay_num):
                delays.append(init_delay + delay_idx * delay_interval)
            self.delay_list = delays

    def get_delay_list(self):
        return self.delay_list

    def set_measure_list(self, measure_list):
        self.measure_list = measure_list

    def get_measure_list(self):
        return self.measure_list

    def set_raw_data(self, input_data):
        self.raw_data = input_data

    def get_raw_data(self):
        return self.raw_data

    def data4d_from_siemens_mosaic_folder(self, input_path):
        dcm_image_list = os.listdir(input_path)
        acquisition_num = len(dcm_image_list)
        for file_idx in range(acquisition_num):
            dcm_img = pd.dcmread(os.path.join(input_path, dcm_image_list[file_idx]))
            acquisition_idx = dcm_img[0x0020, 0x0012].value - 1
            array_img = dcm_img.pixel_array

            if file_idx == 0:
                slice_num = dcm_img[0x0019, 0x100A].value
                side_num = np.ceil(np.sqrt(slice_num))
                a, b = array_img.shape
                row = int(a / side_num)
                col = int(b / side_num)
                array_4d = np.zeros((row, col, slice_num, acquisition_num))

            for slice_idx in range(slice_num):
                shift1 = np.floor(slice_idx / side_num)
                shift2 = int((slice_idx - shift1 * side_num) * row)
                shift1 = int(shift1 * col)
                array_4d[:, :, slice_idx, acquisition_idx] = array_img[shift1:row + shift1, shift2:col + shift2]

        array_4d = np.moveaxis(array_4d, [0, 1, 2, 3], [1, 0, 2, 3])
        self.raw_data = array_4d

    def data4d_from_multi_folder(self, input_path):
        file_list = []
        file_instance_list = []
        file_series_list = []
        for root, dirs, files in os.walk(input_path):
            for file in files:
                file_list.append(os.path.join(root, file))
                img = pd.dcmread(file_list[-1], stop_before_pixels=True)
                file_instance_list.append(int(img[0x0020, 0x0013].value))
                file_series_list.append(int(img[0x0020, 0x0011].value))

        a = int(img[0x0028, 0x0010].value)
        b = int(img[0x0028, 0x0011].value)
        c = np.unique(file_instance_list)
        d = np.unique(file_series_list)

        array_4d = np.zeros((a, b, len(c), len(d)))
        data3d = np.zeros((a, b, len(c)))

        for j in range(len(d)):
            for i in range(len(c)):
                img = pd.dcmread(file_list[i + len(c) * j])
                instance_idx = img[0x0020, 0x0013].value - 1
                data3d[:, :, instance_idx] = img.pixel_array

            array_4d[:, :, :, j] = data3d

        array_4d = np.moveaxis(array_4d, [0, 1, 2, 3], [1, 0, 2, 3])
        self.raw_data = array_4d

    def data4d_from_ge7_folder(self, input_path):
        dcm_image_list = os.listdir(input_path)
        acquisition_num = len(dcm_image_list)
        for file_idx in range(acquisition_num):
            dcm_img = pd.dcmread(os.path.join(input_path, dcm_image_list[file_idx]))
            instance_idx = dcm_img[0x0020, 0x0013].value - 1
            array_img = dcm_img.pixel_array

            if file_idx == 0:
                slice_num = int(dcm_img[0x0020, 0x1002].value / 10)
                images_num = 9
                row, col = array_img.shape
                array_4d = np.zeros((row, col, slice_num, images_num))

            array_4d[:, :, instance_idx % slice_num, int(np.floor(instance_idx / slice_num))] = array_img

        array_4d = np.moveaxis(array_4d, [0, 1, 2, 3], [1, 0, 2, 3])
        self.raw_data = array_4d

    def xydownsample_raw(self, down_sample):
        self.raw_data = self.raw_data[::down_sample, ::down_sample, :, :]

    def gen_m0(self):
        if self.vendor == 'siemens':
            img_m0 = self.raw_data[:, :, :, 0]
            self.img_m0 = np.tile(img_m0[:, :, :, np.newaxis], (1, 1, 1, len(self.delay_list)))
        elif self.vendor == 'ge':
            if self.series_type == '7delay':
                img_m0 = self.raw_data[:, :, :, -1]
                self.img_m0 = np.tile(img_m0[:, :, :, np.newaxis], (1, 1, 1, len(self.delay_list)))
            else:
                _, _, c, _ = self.raw_data.shape
                cutoff = int(c / 2)
                self.img_m0 = self.raw_data[:, :, cutoff:, :]
        elif self.vendor == 'philips':
            _, _, c, _ = self.raw_data.shape
            skip = int((self.measure_list[0] + 1) * 2)
            self.img_m0 = self.raw_data[:, :, ::skip, :]

    def get_m0(self):
        return self.img_m0

    def gen_mask(self, threshold=0):
        mask = np.zeros(self.img_m0.shape)
        if self.vendor == 'siemens':
            self.threshold = threshold
            mask[self.img_m0 >= np.percentile(self.img_m0, 99) * self.threshold] = 1
        elif self.vendor in ['ge', 'philips']:
            for i in range(self.img_m0.shape[3]):
                pd_data = self.img_m0[:, :, :, i]
                self.threshold = 0.5 * np.sum(np.multiply(pd_data, pd_data)) / np.sum(pd_data)
                temp = np.zeros(np.shape(pd_data))
                temp[pd_data >= self.threshold] = 1
                mask[:, :, :, i] = temp
        self.mask = mask

    def get_mask(self):
        return self.mask

    def gen_plds(self):
        if self.vendor == 'siemens':
            a, b, c, _ = self.raw_data.shape
            self.img_plds = np.zeros((a, b, c, len(self.delay_list)))
            cur_idx = 2
            for pld_idx in range(len(self.measure_list)):
                data3d = np.zeros((a, b, c))
                for rep_idx in range(self.measure_list[pld_idx]):
                    control_image = self.raw_data[:, :, :, cur_idx]
                    label_image = self.raw_data[:, :, :, cur_idx + 1]
                    data3d = data3d + label_image.astype(np.float) - control_image.astype(np.float)
                    cur_idx += 2

                data3d = data3d / self.measure_list[pld_idx]
                data3d[data3d < 0] = 0
                self.img_plds[:, :, :, pld_idx] = data3d * self.extra_pld_factor  # TODO check extra factor here

        elif self.vendor == 'ge':
            if self.series_type == '7delay':
                self.img_plds = self.raw_data[:, :, :, :-1]
            else:
                _, _, c, _ = self.raw_data.shape
                cutoff = int(c / 2)
                self.img_plds = self.raw_data[:, :, :cutoff, :]

        elif self.vendor == 'philips':
            a, b, c, d = self.raw_data.shape
            skip = int((self.measure_list[0] + 1) * 2)
            self.img_plds = np.zeros((a, b, int(c / skip), d))
            for rep_idx in range(2, self.measure_list[0], 2):
                control_image = self.raw_data[:, :, range(rep_idx + 1, c, skip), :]
                label_image = self.raw_data[:, :, range(rep_idx, c, skip), :]
                self.img_plds = self.img_plds + label_image.astype(np.single) - control_image.astype(np.single)
            self.img_plds = self.img_plds / self.measure_list[0]

    def get_plds(self):
        return self.img_plds

    def gen_wd(self):
        a, b, c = self.mask[:, :, :, 0].shape
        numerator = np.zeros((a, b, c))
        denominator = np.zeros((a, b, c))
        for i in range(len(self.delay_list)):
            numerator += self.img_plds[:, :, :, i] * self.delay_list[i]
            denominator += self.img_plds[:, :, :, i]

        denominator[denominator == 0] = 1000000
        wd_image = np.divide(numerator, denominator)
        wd_image[self.mask[:, :, :, 0] == 0] = 0
        self.img_wd = wd_image * self.extra_wds_factor

    def get_wd(self):
        return self.img_wd

    def set_att(self, input_img):
        self.img_att = input_img

    def att_from_ge7_folder(self, input_path):

        dcm_image_list = os.listdir(input_path)
        acquisition_num = len(dcm_image_list)
        for file_idx in range(acquisition_num):
            dcm_img = pd.dcmread(os.path.join(input_path, dcm_image_list[file_idx]))
            instance_idx = dcm_img[0x0020, 0x0013].value - 1
            array_img = dcm_img.pixel_array

            if file_idx == 0:
                slice_num = int(dcm_img[0x0020, 0x1002].value / 10)
                row, col = array_img.shape
                array = np.zeros((row, col, slice_num))

            array[:, :, instance_idx] = array_img.astype(np.float32) / 1000

        array = np.moveaxis(array, [0, 1, 2], [1, 0, 2])
        self.img_att = array

    def gen_att_mapping_DAI(self): # Dai (2012)'s method, estimated from graph, doi:10.1002/mrm.23103
        self.img_att = self.img_wd * 1.761 - 1.532
        self.img_att[self.img_att < 0.7] = 0.7
        self.img_att[self.img_att > 3.0] = 3.0
        self.img_att[self.mask[:,:,:,0] == 0] = 0

    def gen_att_mapping(self):
        att_precision = 0.01
        a, b, c = self.img_wd.shape
        attValues = self._generate_att_cache(att_precision)
        av = attValues.tolist()
        att = np.zeros((a, b, c))
        for a_idx in range(a):
            for b_idx in range(b):
                for c_idx in range(c):
                    if self.mask[a_idx][b_idx][c_idx][0] == 1:
                        wdValue = self.img_wd[a_idx][b_idx][c_idx]
                        if not wdValue in av:
                            index2 = binary_search(av, wdValue)
                        else:
                            index2 = av.index(wdValue)
                        att[a_idx][b_idx][c_idx] = self.delay_list[0] + att_precision * index2

        self.img_att = att

    def _generate_att_cache(self, att_precision):
        pul_dur, _, _, r1a = get_balloon_model_parameters()
        output = np.zeros(int((self.delay_list[-1] - self.delay_list[0]) / att_precision) + 1)
        for index in range(len(output)):
            att = self.delay_list[0] + att_precision * index
            plds = np.zeros(len(self.delay_list))
            for delay_idx in range(len(self.delay_list)):
                delay = self.delay_list[delay_idx]
                plds[delay_idx] = np.exp((min([att - delay, 0]) - att) * r1a) - np.exp(0 - (pul_dur + delay) * r1a)
            numerator = 0
            denominator = 0
            for i in range(len(self.delay_list)):
                denominator += plds[i]
                numerator += plds[i] * (self.delay_list[i])
            output[index] = numerator / denominator
        return output

    def gen_cbfs(self, att_correction):
        pul_dur, blc, tag, r1a = get_balloon_model_parameters()
        a, b, c, d = self.img_plds.shape
        cbf_image = np.zeros((a, b, c, d))

        if att_correction == 'corrected':
            att_image = self.img_att
        elif att_correction == 'simple':
            att_image = np.zeros((a, b, c))

        for a_idx in range(a):
            for b_idx in range(b):
                for c_idx in range(c):
                    for d_idx in range(d):
                        if self.mask[a_idx][b_idx][c_idx][d_idx] == 1:
                            m0_vxl = self.img_m0[a_idx][b_idx][c_idx][d_idx]
                            att_vxl = att_image[a_idx][b_idx][c_idx]
                            pld_vxl = self.img_plds[a_idx][b_idx][c_idx][d_idx]
                            cbf_image[a_idx][b_idx][c_idx][d_idx] = 6000 * pld_vxl / m0_vxl * blc * r1a / 2 / tag / (
                                    np.exp((min([0, att_vxl - self.delay_list[d_idx]]) - att_vxl) * r1a)
                                    - np.exp(- r1a * (pul_dur + self.delay_list[d_idx])))

        cbf_image[cbf_image == np.NAN] = 0
        self.img_cbfs = cbf_image

    def gen_cbfs_ge(self):
        self.img_cbfs = np.zeros(self.img_m0.shape)
        for i in range(self.img_m0.shape[3]):
            c_lambda = 0.9
            c_st = 2
            c_t1t = 1.2
            pld = self.delay_list[i]
            c_t1b = 1.4  # 1.5T磁场的参数
            if self.magnetic_field_strength == 3:
                c_t1b = 1.6  # 3.0T磁场的参数
            c_epsilon = 0.6
            c_tau = self.labeling_duration
            nex = self.number_of_averages
            sf = 32  # 在DicomTag里找不到这个信息

            physiology_related = (1 - np.exp(-c_st / c_t1t)) * np.exp(pld / c_t1b) / c_t1b / (
                    1 - np.exp(-c_tau / c_t1b))
            scan_parameter_related = c_lambda / c_epsilon / sf / nex
            constant = 6000 / 2 * physiology_related * scan_parameter_related
            m0 = self.img_m0[:, :, :, i]
            m0[m0 < 1] = 1
            cbf = np.divide(self.img_plds[:, :, :, i], m0) * constant
            cbf[self.mask[:, :, :, i] == 0] = 0
            self.img_cbfs[:, :, :, i] = cbf

    def gen_cbf_tcorr_ge(self):  # 需要已生成ATT图
        a, b, c, d = self.img_m0.shape
        self.img_mcbf = np.zeros([a,b,c])
        att = self.img_att
        for i in range(d):
            c_lambda = 0.9
            c_st = 2
            c_t1t = 1.2
            pld = self.delay_list[i]
            c_t1b = 1.4  # 1.5T磁场的参数
            if self.magnetic_field_strength == 3:
                c_t1b = 1.6  # 3.0T磁场的参数
            c_epsilon = 0.6
            c_tau = self.labeling_duration
            nex = self.number_of_averages
            sf = 32  # 在DicomTag里找不到这个信息

            # 从王老师那里转换过来的，pld = att时，会产生下一行的公式，就跟GE原始的一样了
            # pld_related = np.exp(pld / c_t1b) / (1 - np.exp(-c_tau / c_t1b))
            pld_related = np.exp(pld / c_t1b) / (np.exp(np.minimum(0, pld - att) / c_t1b) - np.exp(-c_tau / c_t1b))
            physio_related = (1 - np.exp(-c_st / c_t1t)) / c_t1b
            scan_param_related = c_lambda / c_epsilon / sf / nex
            constant = 6000 / 2.0 * physio_related * scan_param_related * pld_related
            m0 = self.img_m0[:, :, :, i]
            m0[m0 < 1] = 1
            cbf = np.divide(self.img_plds[:, :, :, i], m0) * constant
            cbf[self.mask[:, :, :, i] == 0] = 0
            self.img_mcbf += cbf
        self.img_mcbf = self.img_mcbf / d
        self.img_mcbf[self.img_mcbf < 0] = 0
        self.img_mcbf[self.img_mcbf > 400] = 0

    def gen_cbfs_ge7delay(self, transit_correction=True):
        self.img_cbfs = np.zeros(self.img_m0.shape)

        extra_factor = 0.88  # 额外的系数

        if len(self.duration_list) == 7:
            lds = self.duration_list
            plds = self.delay_list
        else:
            lds = [0.22, 0.26, 0.30, 0.37, 0.48, 0.68, 1.18]
            plds = [1.0, 1.22, 1.48, 1.78, 2.15, 2.62, 3.32]

        print("lds & plds")
        print(lds)
        print(plds)
        c_lambda = 0.73
        c_t1t = 1.2
        c_t1b = 1.6  # 3.0T磁场的参数
        c_epsilon = 0.6375

        # 计算每一个CBF
        for cbf_ind in range(7):
            lower_mat1 = np.exp(- plds[cbf_ind] / c_t1t)
            lower_mat2 = np.exp(- (plds[cbf_ind] + lds[cbf_ind]) / c_t1t)
            constant_matrix = 6000 / 2 / (lower_mat1 - lower_mat2) * c_lambda / c_t1b / c_epsilon / 32
            m0 = self.img_m0[:, :, :, cbf_ind]
            m0[m0 < 1] = 1
            cbf = np.multiply(np.divide(self.img_plds[:, :, :, cbf_ind], m0), constant_matrix) * extra_factor
            if transit_correction:
                att_factor = 1 + np.power(abs(self.img_att - plds[cbf_ind] + lds[cbf_ind] + 0.3) * 2, 6) / 2
                cbf = np.divide(cbf, att_factor)
            cbf[self.mask[:, :, :, cbf_ind] == 0] = 0
            self.img_cbfs[:, :, :, cbf_ind] = cbf

        # 计算mCBF
        upper_mat = np.exp(self.img_att * transit_correction / c_t1b)
        lower_mat1 = np.exp(- np.maximum(plds[0] - self.img_att * transit_correction, 0) / c_t1t)
        lower_mat2 = np.exp(- np.maximum(plds[6] + lds[6] - self.img_att * transit_correction, 0) / c_t1t)

        constant_matrix = 6000 / 2 * np.divide(upper_mat, lower_mat1 - lower_mat2) * c_lambda / c_t1b / c_epsilon / 32
        cbf = np.multiply(np.divide(self.img_plds[:, :, :, -1], m0), constant_matrix) * extra_factor ** (1 - transit_correction)
        cbf[self.mask[:, :, :, cbf_ind] == 0] = 0
        self.img_mcbf = cbf

    def gen_cbfs_philips(self):
        self.img_cbfs = np.zeros(self.img_m0.shape)
        for i in range(len(self.delay_list)):
            c_lambda = 0.9
            pld = self.delay_list[i]
            c_t1b = 1.35  # 1.5T磁场的参数
            if self.magnetic_field_strength == 3:
                c_t1b = 1.65  # 3.0T磁场的参数
            c_alpha = 0.85
            c_tau = self.labeling_duration

            physiology_related = np.exp(pld / c_t1b) / (1 - np.exp(-c_tau / c_t1b))
            scan_parameter_related = c_lambda / c_alpha
            constant = 6000 / 2 * physiology_related * scan_parameter_related
            m0 = self.img_m0[:, :, :, i]
            m0[m0 < 1] = 1
            cbf = np.divide(self.img_plds[:, :, :, i], m0) * constant
            cbf[self.mask[:, :, :, i] == 0] = 0
            self.img_cbfs[:, :, :, i] = cbf

    def get_cbfs(self):
        return self.img_cbfs

    def gen_diff_cbfs(self):
        self.img_diff_cbfs = self.img_cbfs[:, :, :, 1:] - self.img_cbfs[:, :, :, :-1]

    def get_diff_cbfs(self):
        return self.img_diff_cbfs

    def gen_att_mcbf_fitting(self):
        att_default = self.delay_list[0]
        cbf_default = 80
        fitting_shift = 0.05
        x_arr = self.delay_list
        a, b, c = self.img_m0[:, :, :, 0].shape
        att_image = np.zeros((a, b, c))
        mcbf_image = np.zeros((a, b, c))
        for c_idx in range(c):
            print([c_idx, c])
            for a_idx in range(a):
                for b_idx in range(b):
                    if self.mask[a_idx, b_idx, c_idx, 0] == 1:
                        y_arr = (self.img_plds[a_idx, b_idx, c_idx, :]).flatten() / self.img_m0[a_idx, b_idx, c_idx, 0]
                        try:
                            att_cbf, _ = curve_fit(func, x_arr, y_arr,
                                                   p0=(att_default + fitting_shift, cbf_default),
                                                   maxfev=100)
                        except Exception as e:
                            print(e)
                        else:
                            if att_cbf[0] < 0 or att_cbf[0] > 20:
                                pass
                            else:
                                att_image[a_idx, b_idx, c_idx] = att_cbf[0]
                                mcbf_image[a_idx, b_idx, c_idx] = att_cbf[1]

        self.img_att = att_image * self.extra_att_factor
        self.img_mcbf = mcbf_image * self.extra_cbf_factor

    def get_att(self):
        return self.img_att

    def get_mcbf(self):
        return self.img_mcbf

    def gen_acbv(self):
        sec_to_min = 60
        self.img_acbv = np.multiply(self.img_att, self.img_mcbf) / sec_to_min

    def get_acbv(self):
        return self.img_acbv




import json
import shutil
import subprocess

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase


class DataSorterSiemensPCASL(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)

        delay_num = len(self.delay_time)
        rep_num = len(self.delay_rep)
        if delay_num != rep_num:
            raise DelayNotEqualRepetitionError(delay_num, rep_num)

        if self.data_type == MRI_Type.MRI_TYPE_5Delay_3D_PCASL_SIEMENS and delay_num != 5:
            raise DelayNotEqualSeriesNumberError(delay_num, 5)
        if self.data_type == MRI_Type.MRI_TYPE_1Delay_3D_PCASL_SIEMENS and delay_num != 1:
            raise DelayNotEqualSeriesNumberError(delay_num, 1)


        # PWI图如果被包含进来，排除掉
        self.raw_series = sorted(series_list, key=lambda x: len(x), reverse=True)[0]

    def Sorter(self):
        super(DataSorterSiemensPCASL, self).Sorter()

        # 单延迟PCASL重新计算重复次数
        if self.data_type == MRI_Type.MRI_TYPE_1Delay_3D_PCASL_SIEMENS:
            self.delay_rep = [self.raw_series.load().shape[-1] // 2 - 1]

        java_temp = os.path.join(self.output_root, 'java_temp')
        try:
            reconstructed_dir = self.generate_from_raw_folder(java_temp)
            self._save_head_motion_graph()

            # 计算完后，得到的2D增强Dicom文件，将其重整为2D Dicom；重写方向，位置标签
            def lambda_load_dicom(name):
                mat = pydicom.read_file(os.path.join(reconstructed_dir, name+'.dcm')).pixel_array
                mat = mat.swapaxes(0, 2)
                return mat.swapaxes(0, 1)

            cur_series_number = SeriesNumberStartWith.ASL
            if len(self.delay_time) == 5:
                param_array = None

                file_list = ["m0", "mcbf", "att", "acbv", "cbf1", "cbf2", "cbf3", "cbf4", "cbf5"]
                scaling_list = [1, 0.1, 0.01, 0.001, 0.1, 0.1, 0.1, 0.1, 0.1]
                for f, s, idx in zip(file_list, scaling_list, range(len(file_list))):
                    image = lambda_load_dicom(f.upper())
                    write_dicom_series(image, self.raw_series, f'asl-{f}', cur_series_number,
                                       os.path.join(self.output_root, 'asl', f), slope=s)
                    cur_series_number += 1

                    if param_array is None:
                        param_array = np.zeros(list(image.shape) + [9])
                    param_array[..., idx] = image

                # 后5个是CBF1-5，第3个是ATT
                super().calc_correct_CBF(param_array[...,-5:] * 0.1, self.raw_series, cur_series_number)

            elif len(self.delay_time) == 1:
                image = lambda_load_dicom('M0')
                write_dicom_series(image, self.raw_series, f'asl-m0', cur_series_number,
                                   os.path.join(self.output_root, 'asl', 'm0'), slope=1)
                cur_series_number += 1

                image = lambda_load_dicom('CBF')
                write_dicom_series(image, self.raw_series, f'asl-cbf1', cur_series_number,
                                   os.path.join(self.output_root, 'asl', 'cbf1'), slope=0.1)

                super().calc_correct_CBF(image * 0.1, self.raw_series, cur_series_number + 1)

        except Exception as e:
            logger.error(e)
            raise e
        finally:
            shutil.rmtree(java_temp)

    def generate_from_raw_folder(self, temp_dir):
        # print("层数 = {:d}".format(self.slice_num))
        os.makedirs(os.path.join(temp_dir, "reordered"), exist_ok=True)
        os.makedirs(os.path.join(temp_dir, "reconstructed"), exist_ok=True)
        self._make_json()

        for i, instance in enumerate(self.raw_series):
            instance.save_as(os.path.join(temp_dir, "reordered", "temp_dcm_{:04d}.dcm".format(i)))
        jar_folder = os.path.join(os.getcwd(), "UCLA_java_jar")

        print(jar_folder)
        response = subprocess.run(
            [
                "java", "-jar",
                "-Djava.library.path={}".format(os.path.join(jar_folder, "lib")),
                os.path.join(jar_folder, "ASL_yao_JAVA-master.jar"),
                os.path.join(temp_dir, "param.json")
            ]
        )
        if response.returncode != 0:
            raise Exception("Java程序运行失败，编号: {:d}".format(response.returncode))

        if os.path.isfile(os.path.join(self.output_root, "java_temp", "reconstructed", "finished.txt")):
            logger.debug("完成用Java的图像重建，开始用Python来调整标签")

        return os.path.join(temp_dir, "reconstructed")

    def _save_head_motion_graph(self):
        file_path = os.path.join(self.output_root, "java_temp", "reconstructed", "headMotion.json")
        save_path = os.path.join(self.output_root, "headMotion.png")
        with open(file_path, 'r') as f:
            data = json.load(f)
        # print(data)
        fig = plt.figure()
        matplotlib.rcParams['font.family'] = "simsun"
        fig.set_size_inches(6, 2)
        xdata = np.linspace(1, len(data["controlFD"]), len(data["controlFD"]))
        plt.plot(xdata, data["controlFD"], 'bo-', label='无标记像')
        plt.plot(xdata, data["labelFD"], 'ro-', label='标记像')
        plt.xticks(xdata)
        plt.xlabel("图片序号")
        plt.ylabel("头动(mm)")
        plt.legend()
        plt.tight_layout()
        plt.savefig(save_path, dpi=200)
        plt.figure().clear()
        plt.close()
        plt.cla()
        plt.clf()

    def _make_json(self):
        self.raw_series: SeriesModel
        delay_num = len(self.delay_time)
        if delay_num == 0:
            raise Exception('delay_time is null')
        if delay_num != len(self.delay_rep):
            raise DelayNotEqualRepetitionError(delay_num, len(self.delay_rep))

        param_json_dict = {
            "data_type": '5delay_mosaic' if self.data_type in [MRI_Type.MRI_TYPE_5Delay_3D_PCASL_SIEMENS,
                                                               MRI_Type.MRI_TYPE_1Delay_3D_PCASL_SIEMENS] else '4delay_3d',
            "slice_order": 'ascending', # 此参数已不重要
            "slice_num": self.raw_series.SliceNumber,
            "init_delay": self.delay_time[0],
            "delay_num": delay_num,
            "interval": self.delay_time[1] - self.delay_time[0] if delay_num >= 2 else 0.0,
            "delay_rep": self.delay_rep,
            "load_path": os.path.join(self.output_root, "java_temp", "reordered"),
            "save_path": os.path.join(self.output_root, "java_temp", "reconstructed")}

        with open(os.path.join(self.output_root, "java_temp", "param.json"), "w") as fp:
            json.dump(param_json_dict, fp, indent=4)
        print("json文件保存")

    @staticmethod
    def _3D_to_mosaic(array: np.ndarray):
        width = height = int(np.ceil(array.shape[2]))
        def get_slice(array: np.ndarray, slice: int):
            # 超出范围用 0值填充
            if array.shape[2] > slice:
                return array[:, :, slice]
            else:
                return np.zeros(array[:, :, 0].shape, dtype=array.dtype)

        row_img_vector = []
        for offset in range(width):
            col_index = tuple([i * width + offset for i in range(height)])
            col_img = np.vstack([get_slice(array, c) for c in col_index])
            row_img_vector.append(col_img)

        return np.hstack(row_img_vector)




import SimpleITK as sitk
import pydicom as pd
import numpy as np
import dicom2nifti
import os
import time
import hashlib
import datetime


def create_dcm_series(pixel_array, ref_dcm, save_path, data_type, perfusion_type, series_param, folder_suffix=''):
    input_int = series_param[0]
    slope_adj = series_param[1]
    inter_adj = series_param[2]
    series_tag = data_type+'-'+perfusion_type
    if folder_suffix != '':
        save_folder = data_type + '-' + folder_suffix
    else:
        save_folder = data_type

    reader = sitk.ImageFileReader()
    reader.SetFileName(ref_dcm)
    ref_image = reader.Execute()

    mod_time = time.strftime("%H%M%S")
    mod_date = time.strftime("%Y%m%d")

    writer = sitk.ImageFileWriter()
    writer.KeepOriginalImageUIDOn()

    pixel_array = ((pixel_array + inter_adj) * slope_adj).astype(np.int16)
    for slice_idx in range(np.shape(pixel_array)[2]):
        image = sitk.GetImageFromArray(pixel_array[:, :, slice_idx])
        image = _copy_tags(image, reader, slice_idx+1, mod_time, mod_date, series_tag)
        image = _modify_series_uid(image, reader, series_tag, input_int)
        image = _modify_sop_uid(image, reader, series_tag, slice_idx+1)

        file_target = os.path.join(save_path, save_folder, perfusion_type, "{:04d}.dcm".format(slice_idx))
        writer.SetFileName(file_target)
        writer.Execute(image)
        pd_image = pd.dcmread(file_target)  # TODO, 最好脱离Pydicom的依赖，目前Sitk搞不清楚这一块
        pd_image[0x0028, 0x1052].value = -inter_adj
        pd_image[0x0028, 0x1053].value = 1 / slope_adj
        pd_image = _add_anImage_tag(pd_image, series_tag)
        pd_image.save_as(file_target)


def copy_dcm_slice(image_data, save_path, data_type, perfusion_type, slice_num, override_series_desc=False):
    series_tag = data_type + '-' + perfusion_type
    if override_series_desc:
        image_data[0x0008, 0x103E].value = series_tag
    image_data = _add_anImage_tag(image_data, series_tag)
    image_data = _modify_series_uid(image_data, None, series_tag, None)
    image_data = _modify_sop_uid(image_data, None, series_tag, None)
    image_data.save_as(os.path.join(save_path, data_type, perfusion_type, '{:04d}.dcm'.format(slice_num)))


def _add_anImage_tag(pd_image, series_tag):
    block = pd_image.private_block(0x0099, 'ANIMAGE HEADER', create=True)
    block2 = pd_image.private_block(0x0099, 'Created through AnImageDataSorter', create=True)
    block3 = pd_image.private_block(0x0099, 'Version 1.03.03', create=True)
    block4 = pd_image.private_block(0x0099, datetime.datetime.now().strftime('%c'), create=True)
    block5 = pd_image.private_block(0x0099, series_tag, create=True)
    return pd_image


def _copy_tags(image, reader, instance, mod_time, mod_date, series_desc):
    copy_list = [
        "0010|0010",  # Patient Name
        "0010|0020",  # Patient ID
        "0010|0030",  # Patient Birth Date
        "0020|000d",  # Study Instance UID, for machine consumption
        "0020|0010",  # Study ID, for human consumption
        "0008|0020",  # Study Date
        "0008|0030",  # Study Time
        "0008|0050",  # Accession Number
        "0008|0060",  # Modality
        "0008|0070",  # Manufacturer
        "0008|0080",  # Institution Name
        "0008|1030",  # Study Description
        "0018|0050",  # Slice Thickness
        "0020|0037",  # Image Orientation Patient
        "0028|0030",  # Pixel Spacing
        "0018|0088",  # Slice Spacing
    ]
    for tag_idx in copy_list:
        try:
            image.SetMetaData(tag_idx, reader.GetMetaData(tag_idx))
        except:
            pass

    pixel_spacing = float(reader.GetMetaData("0028|0030").split('\\')[0])
    slice_spacing = float(reader.GetMetaData("0018|0088"))
    image.SetSpacing([pixel_spacing, pixel_spacing, slice_spacing])

    image.SetMetaData("0008|103e", series_desc)
    image.SetMetaData("0008|0031", mod_time)
    image.SetMetaData("0008|0021", mod_date)
    image.SetMetaData("0008|0008", "DERIVED\\SECONDARY\\PROCESSED")
    image.SetMetaData("0020|000e", "1.2.826.0.1.3680043.2.1125." + mod_date + ".1" + mod_time)
    image.SetMetaData("0020|0013", str(instance))
    image.SetMetaData("0020|0032", '0\\0\\'+str(instance * slice_spacing))
    image.SetMetaData("0020|1041", str(instance * slice_spacing))

    return image


def _modify_series_uid(image, reader, filetype, input_int):
    suffix = str(int(hashlib.sha1(filetype.encode("utf-8")).hexdigest(), 16) % (10 ** 8))
    if isinstance(image, pd.Dataset):
        if input_int:
            image[0x0020, 0x0011].value = int(input_int)
        original_string = image[0x0020, 0x000E].value
        cut_length = len(original_string) - len(suffix) - 1
        image[0x0020, 0x000E].value = original_string[:cut_length] + '.' + suffix
    else:
        if input_int:
            image.SetMetaData("0020|0011", str(input_int))
        original_string = reader.GetMetaData("0020|000e")
        cut_length = len(original_string) - len(suffix) - 1
        image.SetMetaData("0020|000e", original_string[:cut_length] + '.' + suffix)

    return image


def _modify_sop_uid(image, reader, filetype, instance_num):
    if isinstance(image, pd.Dataset):
        if not instance_num:
            instance_num = image[0x0020, 0x0013].value
    else:
        if not instance_num:
            instance_num = reader.GetMetaData("0020|0013")
    suffix = str(int(hashlib.sha1(filetype.encode("utf-8")).hexdigest(), 16) % (10 ** 8) + instance_num)

    if isinstance(image, pd.Dataset):
        original_string = image[0x0008, 0x0018].value
        cut_length = len(original_string) - len(suffix) - 1
        image.file_meta.MediaStorageSOPInstanceUID = original_string[:cut_length] + '.' + suffix
        image[0x0008, 0x0018].value = original_string[:cut_length] + '.' + suffix
    else:
        original_string = reader.GetMetaData("0008|0018")
        cut_length = len(original_string) - len(suffix) - 1
        image.SetMetaData("0002|0003", original_string[:cut_length] + '.' + suffix)
        image.SetMetaData("0008|0018", original_string[:cut_length] + '.' + suffix)

    return image


def dcm_to_nii(save_path, data_type, perfusion_type):
    print(os.path.join(save_path, data_type, perfusion_type))
    dicom2nifti.dicom_series_to_nifti(os.path.join(save_path, data_type, perfusion_type),
                                      os.path.join(save_path, data_type, perfusion_type + '.nii.gz'),
                                      reorient_nifti=True)


def iterative_seq_tag_search(obj, tag, input_array):
    return_value = input_array
    if tag in obj:
        return_value.append(obj[tag].value)
    else:
        for elem in obj:
            if elem.VR == "SQ":
                for item in elem:
                    return_value = iterative_seq_tag_search(item, tag, return_value)
    return return_value


if __name__ == '__main__':
    folder_path = r'C:\Users\zeyao\Desktop\dicom test\input'
    folder_path2 = r'C:\Users\zeyao\Desktop\dicom test\output'
    ref_dicom = r'C:\Users\zeyao\Desktop\dicom test\input\MR000000.dcm'
    array = np.zeros((128, 128, 72))
    for slices in range(72):
        array[:, :, slices] = pd.dcmread(os.path.join(folder_path, 'MR{:06d}.dcm'.format(slices))).pixel_array

    os.makedirs(os.path.join(folder_path2, 'asl'), exist_ok=True)
    os.makedirs(os.path.join(folder_path2, 'asl', 'm0'), exist_ok=True)
    series_param = [1, 0.001, 50]
    create_dcm_series(array, ref_dicom, folder_path2, 'asl', 'm0', series_param)
    dcm_to_nii(folder_path2, 'asl', 'm0')



from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase
from .errors import *


class DataSorterPhilipsPCASL(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
        self.M0:SeriesModel
        self.PWPD = [] # 灌注图
        self.CBF = [] # CBF参数图

        # 区别M0与灌注图
        for series in self.series_list:
            if not isinstance(series, SeriesModel):
                continue
            elif series.SeriesDescription.lower().find('source') > -1:
                self.PWPD.append(series)
            elif series.SeriesDescription.lower().find('3d') > -1 and series.SeriesDescription.lower().find('pcasl') > -1:
                self.CBF.append(series)

    def Sorter(self):
        super(DataSorterPhilipsPCASL, self).Sorter()

        delay_count = len(self.delay_time)
        cbf_count = len(self.CBF)
        pwpd_count = len(self.PWPD)
        if cbf_count == 0 and pwpd_count == 0:
            raise NotFoundSeriesError()
        elif cbf_count == 0:
            logger.error('未发现CBF图')
            raise MissingSequenceError('ASL')
        elif pwpd_count == 0:
            logger.warning('未发现PWPD，可能影响配准的效果')

        if cbf_count != delay_count:
            # raise DelayNotEqualSeriesNumberError(delay_count, cbf_count)
            logger.warning(f'输入的PLD长度与实际CBF数量不符, 将补齐/裁切')
            if delay_count < cbf_count:
                self.delay_time += ['undefined'] * (cbf_count - delay_count)

        logger.info(f'使用的PLD为{self.delay_time}')
        logger.info(f'飞利浦{delay_count}延迟数据')

        # 检测多延迟还是单延迟的多次重复
        self.PWPD.sort(key=lambda x: x.SeriesNumber)
        if pwpd_count > cbf_count:
            merge_PWPD = []
            # NumberOfTemporalPositions 此字段为飞利浦重复次数
            NumberOfTemporalPositions = self.PWPD[0].get_tag((0x0020, 0x0105))
            cur_number = -1
            for i, p in enumerate(self.PWPD):

                # TemporalPositionIdentifier 此字段为飞利浦重复第一次，校验是否缺失
                TemporalPositionIdentifier = p.get_tag((0x0020, 0x0100))
                if TemporalPositionIdentifier == 1:
                    merge_PWPD.append([p])
                    cur_number += 1
                else:
                    if i % NumberOfTemporalPositions + 1 != TemporalPositionIdentifier:
                        raise MissingSequenceError(f'NumberOfTemporalPositions {i + 1}')
                    merge_PWPD[cur_number].append(p)
            self.PWPD = merge_PWPD

        if len(self.PWPD) == 0:
            self.M0 = self.CBF
        elif len(self.PWPD) != cbf_count:
            raise DelayNotEqualSeriesNumberError(delay_count, len(self.PWPD))
        else:
            if type(self.PWPD[0]) != SeriesModel:
                self.M0 = self.PWPD[0]
            else:
                self.M0 = self.PWPD
        self.generate_cbf_3d_pcasl()

    def generate_cbf_3d_pcasl(self):
        m0_image = np.concatenate([i.load() for i in self.M0], axis=3)
        m0_image = m0_image[...,0]      # todo 第一个并不一定是M0，不过M0不参与计算，不影响配准

        cbf_image = np.concatenate([i.load() for i in self.CBF], axis=3)
        cur_series_number = SeriesNumberStartWith.ASL
        write_dicom_series(m0_image, self.M0[0], 'asl-m0', cur_series_number, os.path.join(self.output_root, 'asl', 'm0'))

        for i in range(len(self.CBF)):
            cur_series_number += 1

            write_dicom_series(cbf_image[...,i], self.CBF[0], f'asl-cbf{i+1}', cur_series_number,
                               os.path.join(self.output_root, 'asl', f'cbf{i+1}'))

        super().calc_correct_CBF(cbf_image, self.M0[0], cur_series_number + 1)

    def _gen_mask(self, input_image):
        threshold = 0.5 * np.sum(np.multiply(input_image, input_image)) / np.sum(input_image)
        output_image = np.zeros(np.shape(input_image))
        output_image[input_image >= threshold] = 1
        return output_image



from enum import Enum
from typing import List

from .errors import UnSupportDataTypeError, UnidentifiedOrientationError

UnKnown = 'unknown'

class MANUFACTURER(Enum):
    kMANUFACTURER_SIEMENS = 'SIEMENS'
    kMANUFACTURER_PHILIPS = 'PHILIPS'
    kMANUFACTURER_GE = 'GE MEDICAL SYSTEMS'
    kMANUFACTURER_UIH = 'UIH'

    @staticmethod
    def str2MANUFACTURER(m):
        table = {
            MANUFACTURER.kMANUFACTURER_SIEMENS.value: MANUFACTURER.kMANUFACTURER_SIEMENS,
            MANUFACTURER.kMANUFACTURER_PHILIPS.value: MANUFACTURER.kMANUFACTURER_PHILIPS,
            MANUFACTURER.kMANUFACTURER_GE.value: MANUFACTURER.kMANUFACTURER_GE,
            MANUFACTURER.kMANUFACTURER_UIH.value: MANUFACTURER.kMANUFACTURER_UIH,
        }

        m = m.upper()
        for i in MANUFACTURER:
            if m.find(i.value) > -1:
                return i
        else:
            raise UnSupportDataTypeError(m)

class MODALITY(Enum):
    MODALITY_MR = 'MR'
    MODALITY_CT = 'CT'
    MODALITY_OT = 'OT'

    @staticmethod
    def str2MODALITY(s):
        table = {
            MODALITY.MODALITY_MR.value: MODALITY.MODALITY_MR,
            MODALITY.MODALITY_CT.value: MODALITY.MODALITY_CT,
            MODALITY.MODALITY_OT.value: MODALITY.MODALITY_OT,
        }
        if s in table:
            return table[s]
        else:
            raise UnSupportDataTypeError(s)

class ASL_Type(Enum):
    # MR->ASL
    ASL_PCASL = 'pcasl'
    ASL_PASL = 'pasl'
    ASL_UnKnown = UnKnown

    def startswith(self, s: str):
        self.value.startswith(s)

class DICOM_IMAGE_TYPE(Enum):
    DICOM_IMAGE_TYPE_MOSAIC = 'mosaic'
    DICOM_IMAGE_TYPE_3D = '3d'
    DICOM_IMAGE_TYPE_2D = '2d'
    DICOM_IMAGE_TYPE_UnKnown = UnKnown

class SeriesNumberStartWith():
    ASL = 10000
    T1 = 10010
    T2 = 10020
    T2FLAIR = 10030
    DWI = 10040
    Report = 10090

class MRI_Type(Enum):
    # MR type: AnImage custom
    MRI_TYPE_5Delay_3D_PCASL_SIEMENS = '5delay_PCASL_Siemens'
    MRI_TYPE_1Delay_3D_PCASL_SIEMENS = '1delay_PCASL_Siemens'
    MRI_TYPE_4Delay_3D_PCASL_SIEMENS_Old = '4delay_PCASL_Siemens'
    MRI_TYPE_3D_PASL_SIEMENS = '1delay_PASL_Siemens'
    MRI_TYPE_3D_PCASL_GE = '3D_ASL_GE'
    MRI_TYPE_eASL_PCASL_GE = 'eASL_GE'
    MRI_TYPE_3D_PCASL_PHILIPS = 'PCASL_Philips'
    MRI_TYPE_3D_PASL_UIH = 'PASL_UIH'
    MRI_TYPE_UnKnown = 'Unknown'

    MRI_TYPE_T1 = 'T1'
    MRI_TYPE_T2 = 'T2'
    MRI_TYPE_T2_FLAIR = 'T2_FLAIR'
    MRI_TYPE_DWI = 'DWI'

    # MRI_TYPE_ADC = 'ADC'

class SliceDirection(Enum):
    Axial = 'XY' # 轴状面
    Coronal = 'XZ'  # 冠状面
    Sagittal = 'YZ' # 矢状面

    def iSL(self):
        if self == SliceDirection.Axial:
            return 3
        elif self == SliceDirection.Coronal:
            return 2
        elif self == SliceDirection.Sagittal:
            return 1
        else:
            raise UnidentifiedOrientationError('to iSL')

def affine2Orientation(affine: List[float]) -> SliceDirection:
    if len(affine) != 6:
        raise UnidentifiedOrientationError(affine)
    abs_affine = [abs(i) for i in affine] # 忽略正负数（方向），只关注剖面
    first = abs_affine[:3]
    second = abs_affine[3:]
    a = first.index(max(first))
    b = second.index(max(second))
    if a == 0 and b == 1:
        # [ 1 0 0 ]
        # [ 0 1 0 ] 是轴状面
        return SliceDirection.Axial
    elif a == 0 and b == 2:
        # [ 1 0 0 ]
        # [ 0 0 1 ] 是冠状面
        return SliceDirection.Coronal
    elif a == 1 and b == 2:
        # [ 1 0 0 ]
        # [ 0 0 1 ] 是矢状面
        return SliceDirection.Sagittal
    else:
        raise UnidentifiedOrientationError(affine)


SERIES_TYPE_M0 = 'm0'
SERIES_TYPE_CBF = 'asl-cbf'


from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase
from .errors import *


class DataSorterGEeASL(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
        self.output_root = output_root

        self.RAW = None
        self.ATT = None
        self.CORR_CBF = None
        self.STAN_CBF = None

        # 区别PWPD与CBF图
        for series in self.series_list:
            if not isinstance(series, SeriesModel):
                continue
            seriesDescription = series.SeriesDescription.lower().replace('-', ' ')
            if seriesDescription.find('raw') > -1:
                self.RAW = series
            elif seriesDescription.find('transit delay') > -1:
                self.ATT = series
            elif seriesDescription.find('transit corrected flow') > -1 or seriesDescription.find('transit corrected cbf') > -1:
                self.CORR_CBF = series
            elif seriesDescription.find('cbf') > -1 or seriesDescription.find('flow') > -1:
                self.STAN_CBF = series
            # todo 这里如果出现多组灌注图将只使用最后一个: 需要完善

    def validate_series(self):
        if self.RAW is None or self.ATT is None:
            raise MissingSequenceError('ATT and/or RAW')

    def Sorter(self):
        super(DataSorterGEeASL, self).Sorter()
        self.validate_series()

        raw_data_pwpd = self.RAW.load()
        raw_data_pwpd[raw_data_pwpd < 0] = 0

        # M0
        cur_series_number = SeriesNumberStartWith.ASL
        m0_image = raw_data_pwpd[...,-1]
        write_dicom_series(m0_image, self.RAW, 'asl-m0', cur_series_number, os.path.join(self.output_root, 'asl', 'm0'))

        # mask
        mask_image = self._gen_mask(m0_image)

        # PLD
        plds_image = raw_data_pwpd[...,:-1]

        att_image, (cbfs_image_corr, mcbf_image_corr, cbfs_image_ncorr, mcbf_image_ncorr), acbv_image = self.ge_calc_yao(plds_image, m0_image, mask_image)

        use_self_calc = False # todo 是否使用自己计算的CBF
        is_use_ge_corr = True
        if is_use_ge_corr:
            # CORR_CBF 有可能是MCBF，也有可能是CBF1-7
            if (not use_self_calc) and (self.CORR_CBF is not None) and (self.CORR_CBF.Repetition == 1):
                # 使用提供的MCBF
                mcbf_image_corr = self.CORR_CBF.load()[...,0]
            mcbf_image_corr[mcbf_image_corr < 0] = 0
            cur_series_number += 1
            write_dicom_series(mcbf_image_corr * 10, self.RAW, 'asl-mcbf', cur_series_number,
                               os.path.join(self.output_root, 'asl', 'mcbf'), slope=0.1)

            if (not use_self_calc) and (self.CORR_CBF is not None) and (self.CORR_CBF.Repetition == 7):
                # 使用已提供的结果
                cbfs_image_corr = self.CORR_CBF.load()
            cbfs_image_corr[cbfs_image_corr < 0] = 0
            for i in range(cbfs_image_corr.shape[3]):
                cur_series_number += 1
                write_dicom_series(cbfs_image_corr[...,i] * 10, self.RAW, 'asl-cbf%d'%(i+1), cur_series_number,
                                   os.path.join(self.output_root, 'asl', 'cbf%d'%(i+1)), 0.1)
            # super().calc_correct_CBF(cbfs_image_corr, self.RAW, cur_series_number + 1)

        else:
            # STAN_CBF 同样有可能是MCBF，也有可能是CBF1-7
            if (not use_self_calc) and (self.STAN_CBF is not None) and (self.STAN_CBF.Repetition == 1):
                # 使用提供的MCBF
                mcbf_image_ncorr = self.STAN_CBF.load()[..., 0]
            mcbf_image_ncorr[mcbf_image_ncorr < 0] = 0
            cur_series_number += 1
            write_dicom_series(mcbf_image_ncorr * 10, self.RAW, 'asl-mcbf', cur_series_number,
                               os.path.join(self.output_root, 'asl', 'mcbf'), slope=0.1)

            if (not use_self_calc) and (self.STAN_CBF is not None) and (self.STAN_CBF.Repetition == 7):
                # 使用已提供的结果
                cbfs_image_ncorr = self.STAN_CBF.load()
            cbfs_image_ncorr[cbfs_image_ncorr < 0] = 0
            for i in range(cbfs_image_ncorr.shape[3]):
                cur_series_number += 1
                write_dicom_series(cbfs_image_ncorr[..., i] * 10, self.RAW, 'asl-cbf%d' % (i + 1), cur_series_number,
                                   os.path.join(self.output_root, 'asl', 'cbf%d' % (i + 1)), 0.1)

            # super().calc_correct_CBF(cbfs_image_ncorr, self.RAW, cur_series_number + 1)

        # ATT
        cur_series_number += 1
        write_dicom_series(att_image * 1000, self.RAW, 'asl-att', cur_series_number, os.path.join(self.output_root, 'asl', 'att'), 0.001)

        # ACBV
        cur_series_number += 1
        write_dicom_series(acbv_image * 1000, self.RAW, 'asl-acbv', cur_series_number, os.path.join(self.output_root, 'asl', 'acbv'), 0.001)


    def set_up_aslutil(self):
        pass

    def set_scan_parameter(self, series_model: SeriesModel):
        self.number_of_averages = series_model.get_tag((0x0018, 0x0083), 1)
        self.magnetic_field_strength = series_model.get_tag((0x0018, 0x0087), 1)
        self.labeling_duration = series_model.get_tag((0x0043, 0x10A5), 1500) / 1000

    def ge_calc_yao(self, plds_image: np.ndarray, m0_image: np.ndarray, mask_image: np.ndarray):
        # ATT
        att_image = self.ATT.load()[...,-1]
        att_image /= 1000
        # 非校正CBF
        cbfs_image_ncorr, mcbf_image_ncorr = self.gen_cbfs_ge7delay(att_image, plds_image, m0_image, mask_image, False)

        # 校正后CBF
        cbfs_image_corr, mcbf_image_corr = self.gen_cbfs_ge7delay(att_image, plds_image, m0_image, mask_image, True)

        acbv_image = self.gen_acbv(att_image, mcbf_image_corr)
        return att_image, (cbfs_image_corr, mcbf_image_corr, cbfs_image_ncorr, mcbf_image_ncorr), acbv_image

    def gen_cbfs_ge7delay(self, att_image: np.ndarray, plds_image: np.ndarray, m0_image: np.ndarray, mask_image: np.ndarray, transit_correction=True):
        extra_factor = 0.88  # 额外的系数

        c_lambda = 0.73
        c_t1t = 1.2
        c_t1b = 1.6  # 3.0T磁场的参数
        c_epsilon = 0.6375
        plds = self.delay_time
        lds = self.label_dur

        cbfs_image = np.zeros(list(m0_image.shape) + [7])
        m0_image[m0_image < 1] = 1
        # 计算每一个CBF
        for cbf_ind in range(7):
            lower_mat1 = np.exp(- plds[cbf_ind] / c_t1t)
            lower_mat2 = np.exp(- (plds[cbf_ind] + lds[cbf_ind]) / c_t1t)
            constant_matrix = 6000 / 2 / (lower_mat1 - lower_mat2) * c_lambda / c_t1b / c_epsilon / 32
            cbf = np.multiply(np.divide(plds_image[:, :, :, cbf_ind], m0_image), constant_matrix) * extra_factor
            if transit_correction:
                att_factor = 1 + np.power(abs(att_image - plds[cbf_ind] + lds[cbf_ind] + 0.3) * 2, 6) / 2
                cbf = np.divide(cbf, att_factor)
            cbf[mask_image == 0] = 0
            cbfs_image[:, :, :, cbf_ind] = cbf


        # 计算mCBF
        upper_mat = np.exp(att_image * transit_correction / c_t1b)
        lower_mat1 = np.exp(- np.maximum(plds[0] - att_image * transit_correction, 0) / c_t1t)
        lower_mat2 = np.exp(- np.maximum(plds[6] + lds[6] - att_image * transit_correction, 0) / c_t1t)

        constant_matrix = 6000 / 2 * np.divide(upper_mat, lower_mat1 - lower_mat2) * c_lambda / c_t1b / c_epsilon / 32
        mcbf_image = np.multiply(np.divide(plds_image[:, :, :, -1], m0_image), constant_matrix) * extra_factor ** (
                    1 - transit_correction)
        mcbf_image[mask_image == 0] = 0

        return cbfs_image, mcbf_image

    def gen_acbv(self, att_image: np.ndarray, cbf_image: np.ndarray):
        sec_to_min = 60
        acbv_image = np.multiply(att_image, cbf_image) / sec_to_min
        return acbv_image

    def _gen_mask(self, input_image):
        threshold = 0.5 * np.sum(np.multiply(input_image, input_image)) / np.sum(input_image)
        output_image = np.zeros(np.shape(input_image))
        output_image[input_image >= threshold] = 1
        return output_image



"""
在ge13_data_check里增加检查后延迟标记时间的值以及顺序
之前默认认为标记时间的时间差是0.5秒，文件夹顺序是对的

修改了多余重复的代码用变量代替或用函数归纳
修改了跟Siemens有关的描述 => GE

22年02月06日
姚泽山
"""

import os
import pydicom
import numpy as np
from .AnImageASLUtility import AnImageASLUtility
from . import AnImageDCMUtility as DCMUtil


# GE1-3延迟处理
class AnImageDataSorterGE13:
    def __init__(self):
        self.intr_fldr = ''
        self.delay_num = 0
        self.debugMode = False
        self.raw_data_pwpd = np.zeros((1, 1, 1, 1))
        self.raw_data_cbf = np.zeros((1, 1, 1, 1))
        self.dmst, self.dmsl = 0, 0
        self.dmipp, self.dmiop = [], []
        self.dmrow, self.dmcol, self.dmslc = 0, 0, 0
        self.m0_delay_order, self.m0_delay_list = [], []
        self.cbf_delay_order, self.cbf_delay_list = [], []
        self.descending_order = False
        self.aslutil = AnImageASLUtility()

    def set_up_aslutil(self, seriesm0):
        m0keys = sorted(list(seriesm0[0].keys()))
        image_data = pydicom.dcmread(seriesm0[0][m0keys[1]])
        self.aslutil.set_vendor('ge')
        self.aslutil.set_delay_list(self.m0_delay_list)
        self.aslutil.set_measure_list([1] * self.delay_num)
        self.aslutil.set_scan_parameter(image_data)

    def generate_asl_data(self, seriesm0, seriescbf):
        os.makedirs(os.path.join(self.intr_fldr, 'asl'), exist_ok=True)
        m0keys = sorted(list(seriesm0[self.m0_delay_order[0]].keys()))
        ref_img = seriesm0[self.m0_delay_order[0]][m0keys[1]]

        # 从3DASL图复制当中的M0图
        for series_idx in range(self.delay_num):
            m0keys = sorted(list(seriesm0[self.m0_delay_order[series_idx]].keys()))
            for slice_idx in range(self.dmslc):
                image_data = pydicom.dcmread(seriesm0[self.m0_delay_order[series_idx]][m0keys[slice_idx + 1]])
                self.raw_data_pwpd[:, :, slice_idx, series_idx] = image_data.pixel_array

            if self.descending_order:
                cutting = int(self.dmslc / 2)
                self.raw_data_pwpd[:, :, :cutting, series_idx] = \
                    np.flip(self.raw_data_pwpd[:, :, :cutting, series_idx], axis=2)
                self.raw_data_pwpd[:, :, cutting:, series_idx] = \
                    np.flip(self.raw_data_pwpd[:, :, cutting:, series_idx], axis=2)

        self.raw_data_pwpd[self.raw_data_pwpd < 0] = 0
        self.aslutil.set_raw_data(self.raw_data_pwpd)
        self.aslutil.gen_m0()
        self.aslutil.gen_mask()

        output_array = self.aslutil.get_m0()[:, :, :, 0]
        os.makedirs(os.path.join(self.intr_fldr, 'asl', 'm0'), exist_ok=True)
        series_param = [1, 1, 0]
        DCMUtil.create_dcm_series(output_array, ref_img, self.intr_fldr, 'asl', 'm0', series_param)
        DCMUtil.dcm_to_nii(self.intr_fldr, 'asl', 'm0')

        # 从CBF图复制CBF图，或，从原始数据重建CBF图
        if len(seriescbf) == 0:
            self.aslutil.gen_plds()
            self.aslutil.gen_cbfs_ge()
            output_array = self.aslutil.get_cbfs()
        else:
            for series_idx in range(self.delay_num):
                cbfkeys = sorted(list(seriescbf[self.cbf_delay_order[series_idx]].keys()))
                for slice_idx in range(int(self.dmslc / 2)):
                    image_data = pydicom.dcmread(seriescbf[self.cbf_delay_order[series_idx]][cbfkeys[slice_idx + 1]])
                    array_data = image_data.pixel_array
                    self.raw_data_cbf[:, :, slice_idx, series_idx] = array_data

            if self.descending_order:
                self.raw_data_cbf = self.raw_data_cbf[:, :, ::-1, :]
            output_array = self.raw_data_cbf

        for series_idx in range(self.delay_num):
            os.makedirs(os.path.join(self.intr_fldr, 'asl',
                                     'cbf{:d}'.format(series_idx + 1)), exist_ok=True)
            series_param = [10 + series_idx + 1, 1, 0]
            DCMUtil.create_dcm_series(output_array[:, :, :, series_idx], ref_img, self.intr_fldr, 'asl',
                                      'cbf{:d}'.format(series_idx + 1), series_param)
            DCMUtil.dcm_to_nii(self.intr_fldr, 'asl',
                               'cbf{:d}'.format(series_idx + 1))

        # 若一个延迟以上，生成差异图和3个参数图
        if self.delay_num > 1:
            output_array = output_array[:, :, :, 1:] - output_array[:, :, :, :-1]
            for series_idx in range(1, self.delay_num):
                os.makedirs(os.path.join(self.intr_fldr, 'asl',
                                         'cbf{:d}-{:d}'.format(series_idx + 1, series_idx)), exist_ok=True)
                series_param = [20 + series_idx + 1, 1, 1000]
                DCMUtil.create_dcm_series(output_array[:, :, :, series_idx - 1], ref_img, self.intr_fldr, 'asl',
                                          'cbf{:d}-{:d}'.format(series_idx + 1, series_idx), series_param)
                DCMUtil.dcm_to_nii(self.intr_fldr, 'asl',
                                   'cbf{:d}-{:d}'.format(series_idx + 1, series_idx))

            target_files = ["mcbf", "att", "acbv"]
            scaling = [10, 1000, 1000]
            self.aslutil.gen_plds()
            self.aslutil.gen_wd()
            self.aslutil.gen_att_mapping_DAI()
            self.aslutil.gen_cbf_tcorr_ge()
            self.aslutil.gen_acbv()

            for file_idx in range(len(target_files)):
                os.makedirs(os.path.join(self.intr_fldr, 'asl', '{}'.format(target_files[file_idx])), exist_ok=True)
                if file_idx == 0:
                    output_array = self.aslutil.get_mcbf()
                elif file_idx == 1:
                    output_array = self.aslutil.get_att()
                elif file_idx == 2:
                    output_array = self.aslutil.get_acbv()
                series_param = [30 + file_idx + 1, scaling[file_idx], 0]
                DCMUtil.create_dcm_series(output_array, ref_img, self.intr_fldr, 'asl',
                                          '{}'.format(target_files[file_idx]), series_param)
                DCMUtil.dcm_to_nii(self.intr_fldr, 'asl',
                                   '{}'.format(target_files[file_idx]))

    def ge13_data_check(self, seriesm0, seriescbf):
        image_data = pydicom.dcmread([i for i in seriesm0[0].values()][1], stop_before_pixels=True)
        self.delay_num = len(seriesm0)
        try:
            self.dmslc = image_data[0x0020, 0x1002].value
        except:
            raise Exception('Can not get number of slice!')
        try:
            self.dmrow = image_data[0x0028, 0x0010].value
        except:
            raise Exception('Can not get Row size')
        try:
            self.dmcol = image_data[0x0028, 0x0011].value
        except:
            raise Exception('Can not get Column size')
        try:
            self.dmst = image_data[0x0018, 0x0050].value
        except:
            raise Exception('Can not get SliceThickness')
        try:
            self.dmsl = image_data[0x0020, 0x1041].value
        except:
            raise Exception('Can not get SliceLocation')
        try:
            self.dmipp = image_data[0x0020, 0x0032].value
        except:
            raise Exception('Can not get ImagePositionPatient')
        try:
            self.dmiop = image_data[0x0020, 0x0037].value
        except:
            raise Exception('Can not get ImageOrientationPatient')

        temp_img1 = pydicom.dcmread(seriesm0[0][2], stop_before_pixels=True)
        temp_img2 = pydicom.dcmread(seriesm0[0][3], stop_before_pixels=True)
        self.descending_order = ((temp_img2["00201041"].value - temp_img1["00201041"].value) < 0)

        self.raw_data_pwpd = np.zeros((self.dmrow, self.dmcol, self.dmslc, self.delay_num))
        self.raw_data_cbf = np.zeros((self.dmrow, self.dmcol, int(self.dmslc / 2), self.delay_num))

        if seriesm0:  # 查看PWPD图像的后延迟标记时间以及顺序
            m0_delay_list = []
            for series_idx in seriesm0:
                m0_delay_list.append(series_idx[0][1] / 1000)
            self.m0_delay_order = sorted(range(len(m0_delay_list)), key=lambda k: m0_delay_list[k])
            self.m0_delay_list = sorted(m0_delay_list)

        if seriescbf:  # 查看CBF图像的后延迟标记时间以及顺序
            cbf_delay_list = []
            for series_idx in seriescbf:
                cbf_delay_list.append(series_idx[0][1] / 1000)
            self.cbf_delay_order = sorted(range(len(cbf_delay_list)), key=lambda k: cbf_delay_list[k])
            self.cbf_delay_list = sorted(cbf_delay_list)

    def Sorter(self, intermediate_folder, series, debugMode=False):
        self.intr_fldr = intermediate_folder
        self.debugMode = debugMode

        seriesm0 = series["PWPD"]
        seriescbf = {}
        if "CBF" in series.keys():
            seriescbf = series["CBF"]

        self.ge13_data_check(seriesm0, seriescbf)
        print(self.cbf_delay_list)
        self.set_up_aslutil(seriesm0)
        self.generate_asl_data(seriesm0, seriescbf)



"""
应用JAVA程序来做空间矫正和血流灌注重建的处理西门子单延迟、5延迟数据版本
2021-11-14 模块建立
姚泽山
"""

import os
import shutil
import subprocess
import json

import matplotlib
import pydicom
import SimpleITK as sitk
import numpy as np
import math
from . import AnImageDataSorterUtility as SorterUtil
from . import AnImageDCMUtility as DCMUtil
from matplotlib import pyplot as plt


# 西门子1，5延迟处理
class AnImageDataSorterSiemens5:

    def __init__(self):
        self.intermediate_folder = ''
        self.data_type = ""
        self.slice_order = ""
        self.slice_num = 0
        self.init_delay = 0
        self.delay_num = 0
        self.interval = 0
        self.delay_rep = 0
        self.debugMode = False
        self.formatpack = []
        self.row = 0
        self.col = 0
        self.reader_ref = []
        self.writer_cur = []
        self.ref_image_data = []
        self.temp_value = []

    def generate_from_raw_folder(self, seriesm0):
        # print("层数 = {:d}".format(self.slice_num))
        os.makedirs(os.path.join(self.intermediate_folder, "java_temp", "reordered"), exist_ok=True)
        os.makedirs(os.path.join(self.intermediate_folder, "java_temp", "reconstructed"), exist_ok=True)
        self._make_json()

        for i in range(0, len(seriesm0[0]) - 1, 1):
            index = 1 + i * self.slice_num
            # print(seriesm0[0][index])
            shutil.copyfile(seriesm0[0][index], os.path.join(self.intermediate_folder, "java_temp",
                                                          "reordered", "temp_dcm_{:04d}.dcm".format(i)))
        jar_folder = os.path.join(os.getcwd(), "UCLA_java_jar")

        print(jar_folder)
        response = subprocess.run(
            [
                "java", "-jar",
                "-Djava.library.path={}".format(os.path.join(jar_folder, "lib")),
                os.path.join(jar_folder, "ASL_yao_JAVA-master.jar"),
                os.path.join(self.intermediate_folder, "java_temp", "param.json")
            ]
        )
        if response.returncode != 0:
            raise Exception("Java程序运行失败，编号: {:d}".format(response.returncode))

        if os.path.isfile(os.path.join(self.intermediate_folder, "java_temp", "reconstructed", "finished.txt")):
            print("完成用Java的图像重建，开始用Python来调整标签")

        if self.delay_num == 1:
            file_list = ["m0", "cbf"]
            scaling_list = [1, 0.1]
            self._3d_dcm_to_2d_and_nii(file_list[0], file_list[0], 20, scaling_list[0])
            self._3d_dcm_to_2d_and_nii(file_list[1], "asl-" + file_list[1]+"1", 21, scaling_list[1])

        if self.delay_num == 5:  # TODO 增加CBF2-1的输出
            file_list = ["m0", "mcbf", "att", "acbv", "cbf1", "cbf2", "cbf3", "cbf4", "cbf5"]
            scaling_list = [1, 0.1, 0.01, 0.001, 0.1, 0.1, 0.1, 0.1, 0.1]
            self._3d_dcm_to_2d_and_nii(file_list[0], file_list[0], 20, scaling_list[0])
            self._save_head_motion_graph()
            for file_idx in range(1, len(file_list)):
                self._3d_dcm_to_2d_and_nii(file_list[file_idx], "asl-"+file_list[file_idx],
                                           file_idx+20, scaling_list[file_idx])

        print("删除java_temp文件夹，开始自动化所CMD应用")
        shutil.rmtree(os.path.join(self.intermediate_folder, "java_temp"))

    def _save_head_motion_graph(self):
        file_path = os.path.join(self.intermediate_folder, "java_temp", "reconstructed", "headMotion.json")
        save_path = os.path.join(self.intermediate_folder, "headMotion.png")
        with open(file_path, 'r') as f:
            data = json.load(f)
        # print(data)
        fig = plt.figure()
        matplotlib.rcParams['font.family'] = "simsun"
        fig.set_size_inches(6, 2)
        xdata = np.linspace(1, len(data["controlFD"]), len(data["controlFD"]))
        plt.plot(xdata, data["controlFD"], 'bo-', label='无标记像')
        plt.plot(xdata, data["labelFD"], 'ro-', label='标记像')
        plt.xticks(xdata)
        plt.xlabel("图片序号")
        plt.ylabel("头动(mm)")
        plt.legend()
        plt.tight_layout()
        plt.savefig(save_path, dpi=200)
        plt.figure().clear()
        plt.close()
        plt.cla()
        plt.clf()

    def _make_json(self):
        param_json_dict = {"data_type": self.data_type,
                           "slice_order": self.slice_order,
                           "slice_num": self.slice_num,
                           "init_delay": self.init_delay,
                           "delay_num": self.delay_num,
                           "interval": self.interval,
                           "delay_rep": self.delay_rep,
                           "load_path": os.path.join(self.intermediate_folder, "java_temp", "reordered"),
                           "save_path": os.path.join(self.intermediate_folder, "java_temp", "reconstructed")}

        with open(os.path.join(self.intermediate_folder, "java_temp", "param.json"), "w") as fp:
            json.dump(param_json_dict, fp, indent=4)
        print("json文件保存")

    def _3d_dcm_to_2d_and_nii(self, input_name, series_description, series_id, scaling):

        output = DCMUtil.iterative_seq_tag_search(self.ref_image_data, "00200032", [])
        img_pos_pat = output[0]
        output = DCMUtil.iterative_seq_tag_search(self.ref_image_data, "00201041", [])
        if len(output) == 0:
            slice_loc = float(img_pos_pat[2])
        else:
            slice_loc = output[0]
        output = DCMUtil.iterative_seq_tag_search(self.ref_image_data, "00180088", [])
        spacing = float(output[0])
        output = DCMUtil.iterative_seq_tag_search(self.ref_image_data, "00280030", [])
        px_spacing = output[0]
        if input_name == "cbf":
            input_name_mod = input_name+"1"
        else:
            input_name_mod = input_name

        os.makedirs(os.path.join(self.intermediate_folder, 'asl', input_name_mod), exist_ok=True)
        print("读取 "+input_name)
        image = pydicom.dcmread(os.path.join(self.intermediate_folder, "java_temp",
                                             "reconstructed", input_name.upper()+".dcm"))
        arr = image.pixel_array
        self.writer = sitk.ImageFileWriter()
        self.writer.KeepOriginalImageUIDOn()

        skip_tag_list = []
        for i in range(arr.shape[0]):
            image = sitk.GetImageFromArray(arr[i, :, :].astype(np.uint16))

            copy_list = [
                "0010|0010",  # Patient Name
                "0010|0020",  # Patient ID
                "0010|0030",  # Patient Birth Date
                "0020|000d",  # Study Instance UID, for machine consumption
                "0020|0010",  # Study ID, for human consumption
                "0008|0020",  # Study Date
                "0008|0030",  # Study Time
                "0008|0050",  # Accession Number
                "0008|0060",  # Modality
                "0008|0070",  # Manufacturer
                "0008|0080",  # Institution Name
                "0008|1030",  # Study Description
                "0018|0050",  # Slice Thickness
                "0020|0037",  # Image Orientation Patient
                "0028|0030",  # Pixel Spacing
                "0018|0088",  # Slice Spacing
                "0020|000e",  # Series UID
            ]

            for elem in copy_list:
                try:
                    image.SetMetaData(elem, str(self.ref_image_data[elem.replace("|", "")].value))
                except Exception:
                    if elem not in skip_tag_list:
                        skip_tag_list.append(elem)
                        print("跳过了标签：{}".format(elem))

            # image.SetMetaData("0008|0008", "ORIGINAL\\SECONDARY\\ASL\\ANIMAGE")
            image.SetMetaData("0008|0050", str(self.slice_num))
            image.SetMetaData("0020|0013", str(i + 1))
            image.SetMetaData("0020|0037", "1\\0\\0\\0\\1\\0")  # 强制三维图像的垂直性
            image.SetMetaData("0020|0032", str(img_pos_pat[0]) + "\\" + str(img_pos_pat[1]) + "\\" +
                              str(float(img_pos_pat[2]) + i * spacing - (slice_loc >= 0) * 40 * spacing))
            image.SetMetaData("0020|1041", str(slice_loc + i * spacing - (slice_loc >= 0) * 40 * spacing))
            save_path = os.path.join(self.intermediate_folder, "asl", input_name_mod, "{:04}.dcm".format(i + 1))

            self.writer.SetFileName(save_path)
            self.writer.Execute(image)

            temp_image = pydicom.dcmread(save_path)
            temp_image = SorterUtil.add_AnImage_tag(temp_image, series_description)
            temp_image = SorterUtil.modify_series_uid(temp_image, series_description, series_id)
            temp_image[0x0018,0x0088].value = spacing
            temp_image[0x0028,0x0030].value = [float(px_spacing[0]), float(px_spacing[1])]
            if [0x0028, 0x1052] in temp_image:
                temp_image[0x0028, 0x1052].value = 0
            else:
                temp_image.add_new([0x0028, 0x1052], "DS", 0)
            if [0x0028, 0x1053] in temp_image:
                temp_image[0x0028, 0x1053].value = scaling
            else:
                temp_image.add_new([0x0028, 0x1053], "DS", scaling)
            temp_image.save_as(save_path, write_like_original=False)

        DCMUtil.dcm_to_nii(self.intermediate_folder, 'asl', input_name_mod)

    def parse_m0_from_raw_folder(self, seriesm0):
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'm0'), exist_ok=True)
        try:
            image_data = pydicom.dcmread(seriesm0[1])
        except:
            image_data = pydicom.dcmread([i for i in seriesm0.values()][1])
        # m0 = image_data.pixel_array
        self.row = image_data[0x0028, 0x0010].value
        self.col = image_data[0x0028, 0x0011].value
        # mask = np.zeros((self.row, self.col))
        # threshold = 0.15
        # mask[m0 < m0.max() * threshold] = 1
        if "MOSAIC" in image_data[0x0008, 0x0008].value:
            slice_num = image_data[0x0019, 0x100A].value
        else:
            slice_num = image_data[0x0028, 0x0008].value
        sn = math.ceil(np.sqrt(slice_num))
        dmrow = image_data[0x0028, 0x0010].value // sn
        siemensst = image_data[0x0018, 0x0050].value
        siemenssl = image_data[0x0020, 0x1041].value
        siemensipp = image_data[0x0020, 0x0032].value
        siemensiop = [1, 0, 0, 0, 1, 0]
        self.formatpack = [slice_num, sn, dmrow, siemensst, siemenssl, siemensipp, siemensiop]
        SorterUtil.demosaic(image_data, image_data.pixel_array, 'asl-m0', 'm0', self.intermediate_folder, self.formatpack)

    def siemens_data_check(self, seriesm0):
        try:
            self.ref_image_data = pydicom.dcmread(seriesm0[0][1])
        except:
            self.ref_image_data = pydicom.dcmread([i for i in seriesm0[0].values()][1])

        if "MOSAIC" in self.ref_image_data[0x0008, 0x0008].value:
            self.data_type = "5delay_mosaic"
            try:
                self.slice_num = int(float(self.ref_image_data[0x0019, 0x100a].value))
            except:
                raise Exception('slice number not found')
            # if float(self.ref_image_data[0x0020,0x1041].value) >= 0:
            #     self.slice_order = "descending"
            # elif float(self.ref_image_data[0x0020,0x1041].value) < 0:
            #     self.slice_order = "ascending"
            self.slice_order = "ascending"
        else:
            self.data_type = "5delay_3d"
            self.slice_num = self.ref_image_data[0x0028, 0x0008].value
            print("该pCASL序列是3D文件（非mosaic）")

    def Sorter(self, intermediate_folder, seriesm0, delay_time, delay_rep, debugMode=False):
        self.intermediate_folder = intermediate_folder
        self.delay_rep = delay_rep
        self.delay_num = len(delay_time)
        self.init_delay = float(delay_time[0])  # 导进来的init_delay其实是每一个delay的时间点
        if len(delay_time) > 1:
            self.interval = float(delay_time[1] - delay_time[0])
        else:
            self.interval = float(0)
        self.debugMode = debugMode
        os.makedirs(os.path.join(self.intermediate_folder, 'asl'), exist_ok=True)
        self.siemens_data_check(seriesm0)
        self.generate_from_raw_folder(seriesm0)



from scipy.ndimage import gaussian_filter

from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase
from .errors import *


class DataSorterSiemensPASL(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, extra_factor, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
        if extra_factor is None or extra_factor == 0:
            raise Exception('extra_factor 未设置')
        self.extra_factor = 32 / extra_factor
        self.M0 = None
        self.PWPD = [] # 灌注图
        self.CBF = None # CBF参数图

        # 区别M0与灌注图
        for series in self.series_list:
            series_description = series.SeriesDescription.lower()
            if not isinstance(series, SeriesModel):
                continue
            if series_description.find('m0') > -1 or series_description.find('mo') > -1:
                self.M0 = series
            elif series_description.find('asl_3d_tra') > -1 or series_description.find('asl_pasl') > -1:
                self.PWPD.append(series)

        delay_count = len(self.delay_time)
        pwpd_count = len(self.PWPD)

        # 判断是否缺少某些序列
        if pwpd_count == 0:
            logger.error('未发现灌注图')
            raise MissingSequenceError('ASL')
        elif delay_count != pwpd_count:
            raise DelayNotEqualSeriesNumberError(delay_count, pwpd_count)
        if self.M0 is None:
            logger.warning('未发现M0，可能影像数值计算的准确性')
            self.M0 = self.PWPD[0]

        # 不同的重复次数存在于不同序列需要重新排序
        self.PWPD.sort(key=lambda x: (x.SeriesNumber, x[0].InstanceNumber))

    def Sorter(self):
        super(DataSorterSiemensPASL, self).Sorter()
        if self.PWPD[0].SeriesDescription.lower().find('asl_3d_tra_p2') > -1:
            self.generate_cbf_3d_pcasl()
        else:
            self.generate_cbf_3d_pasl()

    def generate_cbf_3d_pcasl(self):
        self.M0: SeriesModel
        m0_image = self.M0.load()
        # 多张M0只有第一张有效
        if len(m0_image.shape) == 4:
            m0_image = m0_image[:,:,:,0]
        m0_image *= self.extra_factor

        cur_series_number = SeriesNumberStartWith.ASL
        write_dicom_series(m0_image, self.M0, 'asl-m0', cur_series_number, os.path.join(self.output_root, 'asl', 'm0'))

        for i, pwpd in enumerate(self.PWPD):
            cbf_pcasl_calc = self.pcasl_calc_cbf(pwpd.load(), m0_image, self.delay_time[i], self.label_dur[i])
            cur_series_number += 1
            write_dicom_series(cbf_pcasl_calc * 10, pwpd, f'asl-cbf{i+1}', cur_series_number,
                               os.path.join(self.output_root, 'asl', f'cbf{i+1}'), slope=0.1)

    def pcasl_calc_cbf(self, pwpd_image: np.ndarray, m0_image: np.ndarray, delay_time, label_dur):
        control = pwpd_image[: ,: ,: ,0::2] # 控制像
        label = pwpd_image[:, :, :, 1::2] # 标记像
        repetitions = control.shape[-1]
        pwi = np.zeros(m0_image.shape)
        sub = control - label
        for i in range(repetitions):
            pwi += sub[:,:,:,i]
        pwi /= repetitions

        m0_image[m0_image < 1] = 1
        mask = self._gen_mask(m0_image)

        # calculating using simple model
        u_cnv = 6000  # units conversion from mL/g/s to mL/100g/min
        bb_p = 0.9  # mL/g, blood-brain partition coefficient
        t_inv = delay_time  # sec, post-labeling delay
        t1_b = 1.65  # sec, relaxation of blood at 3.0T
        lab_eff = 0.85  # labeling efficiency for pCASL
        l_dur = label_dur  # sec, labeling duration
        NEX = 32
        upper_param = u_cnv * bb_p * np.exp(t_inv / t1_b)
        lower_param = 2 * lab_eff * t1_b * (1 - np.exp(-l_dur / t1_b))
        cbf_pcasl_calc = np.divide(pwi, m0_image) / NEX * upper_param / lower_param

        cbf_pcasl_calc[pwi < 8] = 0
        cbf_pcasl_calc = np.multiply(cbf_pcasl_calc, mask)
        cbf_pcasl_calc = cbf_pcasl_calc.astype(np.int16)
        return cbf_pcasl_calc

    def generate_cbf_3d_pasl(self):
        self.M0: SeriesModel
        m0_image = self.M0.load()
        # 多张M0只有第一张有效
        m0_image = m0_image[:,:,:,0]
        m0_image *= self.extra_factor
        cur_series_number = SeriesNumberStartWith.ASL
        write_dicom_series(m0_image, self.M0, 'asl-m0', cur_series_number, os.path.join(self.output_root, 'asl', 'm0'))

        for i, pwpd in enumerate(self.PWPD):
            cbf_pasl_calc = self.pasl_calc_cbf(pwpd.load(), m0_image, self.delay_time[i], self.label_dur[i])
            cur_series_number += 1
            write_dicom_series(cbf_pasl_calc * 10, pwpd, f'asl-cbf{i+1}', cur_series_number,
                               os.path.join(self.output_root, 'asl', f'cbf{i+1}'), slope=0.1)

    def pasl_calc_cbf(self, pwpd_image: np.ndarray, m0_image: np.ndarray, delay_time, label_dur):
        control = pwpd_image[:, :, :, 0::2]  # 控制像
        label = pwpd_image[:, :, :, 1::2]  # 标记像
        repetitions = control.shape[-1]
        pwi = np.zeros(m0_image.shape)
        sub = control - label
        for i in range(repetitions):
            pwi += sub[:, :, :, i]
        pwi /= repetitions

        # calculating using simple model
        u_cnv = 6000  # units conversion from mL/g/s to mL/100g/min
        bb_p = 0.9  # mL/g, blood-brain partition coefficient
        t_inv = delay_time  # sec, inversion time
        t1_b = 1.65  # sec, relaxation of blood at 3.0T
        lab_eff = 0.98  # labeling efficiency for PASL
        l_dur = label_dur  # sec, labeling duration

        m0_image[m0_image < 1] = 1
        mask = self._gen_mask(m0_image)

        check = np.multiply(mask, pwi)

        if sum(check.flatten()) < 0:
            pwi = -pwi
        pwi[pwi < 0] = 0
        upper_param = u_cnv * bb_p * np.exp(t_inv / t1_b)
        lower_param = 2 * lab_eff * l_dur
        cbf_pasl_calc = pwi / m0_image * upper_param / lower_param
        cbf_pasl_calc = np.multiply(cbf_pasl_calc, mask)
        cbf_pasl_calc = gaussian_filter(cbf_pasl_calc, 0.7)
        cbf_pasl_calc = cbf_pasl_calc.astype(np.int16)
        return cbf_pasl_calc

    def _gen_mask(self, input_image):
        threshold = 0.5 * np.sum(np.multiply(input_image, input_image)) / np.sum(input_image)
        output_image = np.zeros(np.shape(input_image))
        output_image[input_image >= threshold] = 1
        return output_image


import numpy as np
import os
import datetime
from pydicom.dataset import FileMetaDataset
import SimpleITK as sitk
import pydicom as pd
import dicom2nifti
import hashlib


## 杂项
def demosaic(input_image, data_array, dcm_tag, save_name, intermediate_folder, formatpack):
    ## formatpack0=slice_num
    #  formatpack1=sn
    #  formatpack2=dmrow
    #  formatpack3=siemensst
    #  formatpack4=siemenssl
    #  formatpack5=siemensipp
    #  formatpack6=siemensiop
    a = 0
    v = np.zeros((formatpack[0], formatpack[2], formatpack[2]))
    sid3 = input_image[0x0008, 0x0008].value
    sid4 = []
    for i in sid3:
        if i != 'MOSAIC':
            sid4.append(i)

    bitsStored = len(bin(int(data_array.max()))) - 2
    for i in np.vsplit(data_array, formatpack[1]):
        for j in np.hsplit(i, formatpack[1]):
            if a >= formatpack[0]:
                break
            jf = np.squeeze(j)
            v[a, :, :] = jf
            input_image[0x0028, 0x0101].value = bitsStored
            input_image[0x0028, 0x0102].value = bitsStored - 1

            input_image.Columns, input_image.Rows = formatpack[2], formatpack[2]
            input_image = get_and_set_bytetype(input_image, jf)
            input_image[0x0020, 0x0013].value = a + 1
            input_image[0x0020, 0x1041].value = formatpack[4] + a * formatpack[3]
            tempipp = formatpack[5]
            tempipp[2] = '{:.4f}'.format(input_image[0x0020, 0x1041].value)
            input_image[0x0020, 0x0032].value = tempipp
            input_image[0x0008, 0x0070].value = 'SIEMENS'
            input_image[0x0020, 0x0037].value = formatpack[6]
            input_image.AcquisitionNumber = a + 1
            sid = str(input_image.SourceImageSequence[a].ReferencedSOPInstanceUID)
            # print(input_image.file_meta)
            input_image[0x0008, 0x0008].value = sid4
            input_image[0x0008, 0x0018].value = sid
            input_image = add_AnImage_tag(input_image, dcm_tag)
            input_image.file_meta.MediaStorageSOPInstanceUID = sid
            modify_series_uid(input_image, 'm0', False)
            input_image.save_as(os.path.join(intermediate_folder, 'asl', save_name, '{:04d}.dcm'.format(a + 1)),
                                write_like_original=False)
            a += 1

    dicom2nifti.dicom_series_to_nifti(os.path.join(intermediate_folder,
                                                   'asl', save_name),
                                      os.path.join(intermediate_folder,
                                                   'asl', save_name + '.nii.gz'), reorient_nifti=False)


def to_mosaic(array_3d):
    [a,b,c] = np.shape(array_3d)
    factor = np.ceil(np.sqrt(c))
    array_2d = np.zeros((int(a * factor), int(b * factor)))
    for slice_idx in range(c):
        a_idx = int(np.floor(slice_idx / factor)) * a
        b_idx = int(np.remainder(slice_idx, factor)) * b
        array_2d[a_idx:a_idx+a, b_idx:b_idx+b] = array_3d[:, :, slice_idx]
    return array_2d


def demosaic_sitk(orig_input_image, data_array, tag, intermediate_folder, formatpack):
    writer = sitk.ImageFileWriter()
    a = 0
    v = np.zeros((formatpack[0], formatpack[2], formatpack[2]))
    sid3 = orig_input_image[0x0008, 0x0008].value
    sid4 = []
    for i in sid3:
        if i != 'MOSAIC':
            sid4.append(i)
    for i in np.vsplit(data_array, formatpack[1]):
        for j in np.hsplit(i, formatpack[1]):
            if a >= formatpack[0]:
                break
            jf = np.squeeze(j)
            sitk_img = sitk.GetImageFromArray(jf.astype(np.uint16))
            writer.SetFileName("temp.dcm")
            writer.Execute(sitk_img)

            input_image = pd.dcmread("temp.dcm")
            for element in orig_input_image:
                if np.floor(int(element.tag) / 16 ** 4) in [0x8, 0x10, 0x18, 0x20, 0x28]:
                    if element.tag in input_image:
                        input_image[element.tag].value = element.value
                    else:
                        input_image.add_new(element.tag, element.VR, element.value)

            input_image.Columns, input_image.Rows = formatpack[2], formatpack[2]
            # input_image = get_and_set_bytetype(input_image, jf)
            input_image[0x0020, 0x0013].value = a + 1
            input_image[0x0020, 0x1041].value = formatpack[4] + a * formatpack[3]
            tempipp = formatpack[5]
            tempipp[2] = '{:.4f}'.format(input_image[0x0020, 0x1041].value)
            input_image[0x0020, 0x0032].value = tempipp
            input_image[0x0008, 0x0070].value = 'SIEMENS'
            input_image[0x0020, 0x0037].value = formatpack[6]
            input_image.AcquisitionNumber = a + 1
            sid = str(input_image.SourceImageSequence[a].ReferencedSOPInstanceUID)
            # print(input_image.file_meta)
            input_image[0x0008, 0x0008].value = sid4
            input_image[0x0008, 0x0018].value = sid
            input_image = add_AnImage_tag(input_image, "asl-"+tag)
            input_image.file_meta.MediaStorageSOPInstanceUID = sid
            modify_series_uid(input_image, tag, None)
            input_image.save_as(os.path.join(intermediate_folder, 'asl', tag, '{:04d}.dcm'.format(a + 1)),
                                write_like_original=False)
            a += 1

    dicom2nifti.dicom_series_to_nifti(os.path.join(intermediate_folder,
                                                   'asl', tag),
                                      os.path.join(intermediate_folder,
                                                   'asl', tag + '.nii.gz'), reorient_nifti=False)


def modify_uid_tag_AnImage(input_image, filetype, instance_num=None, input_int=None):
    input_image = modify_series_uid(input_image, filetype, input_int)
    input_image = modify_sop_uid(input_image, filetype, instance_num)
    input_image = add_AnImage_tag(input_image, filetype)
    return input_image


def modify_series_uid(input_image, filetype, input_int):
    suffix = str(int(hashlib.sha1(filetype.encode("utf-8")).hexdigest(), 16) % (10 ** 8))
    original_string = input_image[0x0020, 0x000E].value
    cut_length = len(original_string) - len(suffix) - 1
    input_image[0x0020, 0x000E].value = original_string[:cut_length] + '.' + suffix
    if input_int:
        input_image[0x0020, 0x0011].value = int(input_int)
    return input_image


def modify_sop_uid(input_image, filetype, instance_num):
    if not instance_num:
        instance_num = input_image[0x0020, 0x0013].value
    suffix = str(int(hashlib.sha1(filetype.encode("utf-8")).hexdigest(), 16) % (10 ** 8) + instance_num)
    original_string = input_image[0x0008, 0x0018].value
    cut_length = len(original_string) - len(suffix) - 1
    input_image.file_meta.MediaStorageSOPInstanceUID = original_string[:cut_length] + '.' + suffix
    input_image[0x0008, 0x0018].value = original_string[:cut_length] + '.' + suffix
    return input_image


def add_AnImage_tag(input_image, filetype):
    if [0x0008, 0x0008] in input_image:
        input_image[0x0008, 0x0008].value = ['DERIVED', 'SECONDARY', 'PROCESSED']
    else:
        input_image.add_new([0x0008, 0x0008], 'CS', ['DERIVED', 'SECONDARY', 'PROCESSED'])
    input_image.remove_private_tags()
    tag = filetype.lower()
    block = input_image.private_block(0x0099, 'ANIMAGE HEADER', create=True)
    block2 = input_image.private_block(0x0099, 'Created through AnImageDataSorter', create=True)
    block3 = input_image.private_block(0x0099, 'Version 1.03.03', create=True)
    block4 = input_image.private_block(0x0099, datetime.datetime.now().strftime('%c'), create=True)
    block5 = input_image.private_block(0x0099, tag, create=True)
    if not ('t1' in filetype or 't2' in filetype):
        if [0x0018, 0x1030] in input_image:
            input_image[0x0018, 0x1030].value = tag
        else:
            input_image.add_new([0x0018, 0x1030], 'LO', tag)
        if [0x0008, 0x103e] in input_image:
            input_image[0x0008, 0x103e].value = tag
        else:
            input_image.add_new([0x0008, 0x103e], 'LO', tag)
    return input_image


def get_and_set_bytetype(dicom_image, num_data):
    bytetype = dicom_image[0x0028, 0x0100].value
    if dicom_image.is_little_endian == False:
        del dicom_image.file_meta
        file_meta = FileMetaDataset()
        file_meta.FileMetaInformationGroupLength = 180
        file_meta.FileMetaInformationVersion = b'\x00\x01'
        file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.4'
        file_meta.MediaStorageSOPInstanceUID = '1.2.840.113619.2.80.2406827568.2282.1597905292.45'
        file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'
        file_meta.ImplementationClassUID = '1.2.40.0.13.1.1'
        file_meta.ImplementationVersionName = 'dcm4che-3.3.7'
        dicom_image.file_meta = file_meta
        dicom_image.is_implicit_VR = False
        dicom_image.is_little_endian = True

    if bytetype == 16:
        dicom_image.PixelData = num_data.astype(np.uint16)
    elif bytetype == 8:
        dicom_image.PixelData = num_data.astype(np.uint8)
    else:
        raise Exception('invalid bytetype!')
    return dicom_image


def parse_dcm_files_add_tag(series, folder, type, name, multiple):
    if multiple == 'no':
        os.makedirs(os.path.join(folder, type), exist_ok=True)
        if name == '':
            tag_name = type
            nifti_name = type
        else:
            os.makedirs(os.path.join(folder, type, name), exist_ok=True)
            tag_name = type+'-'+name
            nifti_name = name
        for slice_idx in range(len(series) - 1):
            image_data = pd.dcmread(series[slice_idx + 1])
            image_data = add_AnImage_tag(image_data, tag_name)
            image_data = modify_series_uid(image_data, tag_name)
            image_data.save_as(os.path.join(folder, type, name,
                                            '{:04d}.dcm'.format(slice_idx + 1)))
        if type == 't1' or type == 't2':
            type = ''
        dcm_to_nii(os.path.join(folder, type), nifti_name)

    elif multiple == 'type':
        for series_id in range(len(series)):
            os.makedirs(os.path.join(folder, type+'_{:d}'.format(series_id + 1)))
            for slice_idx in range(len(series[series_id]) - 1):
                image_data = pd.dcmread(series[series_id][slice_idx + 1])
                image_data = add_AnImage_tag(image_data, type)
                image_data.save_as(os.path.join(folder,
                                                type+'_{:d}'.format(series_id + 1),
                                                '{:04d}.dcm'.format(slice_idx + 1)))


def dcm_to_nii(save_path, data_type, perfusion_type=''):
    if perfusion_type == '':
        save_name = data_type
    else:
        save_name = perfusion_type

    print(os.path.join(save_path, data_type, perfusion_type, save_name + '.nii.gz'))
    dicom2nifti.dicom_series_to_nifti(os.path.join(save_path, data_type, perfusion_type),
                                      os.path.join(save_path, data_type, perfusion_type, save_name + '.nii.gz'),
                                      reorient_nifti=True)



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
  long_description = fh.read()

setuptools.setup(
  name="animage-data_sorter",
  version="1.1.2",
  author="zhaomy",
  author_email="zhaomy@an-image.cn",
  description="dicom sort",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://gitee.com/AnImage-Beijing/data_sorter",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)


"""
西门子的PASL后处理
2022-02-11 模块建立
姚泽山
"""

import os
import pydicom
import SimpleITK as sitk
import numpy as np
import math
from . import AnImageDCMUtility as DCMUtil
from . import AnImageDataSorterUtility as SorterUtil
from scipy.ndimage import gaussian_filter


# 西门子PASL1延迟处理
class AnImageDataSorterSiemensPASL:

    def __init__(self):
        self.intermediate_folder = ''
        self.slice_order = 0
        self.slice_num = 0
        self.delay_time = 0
        self.delay_num = 0
        self.delay_rep = 0
        self.label_dur = 0
        self.debugMode = False
        self.formatpack = []
        self.row = 0
        self.col = 0
        self.reader_ref = []
        self.writer_cur = []
        self.ref_image_data = []

    def generate_cbf_2d_pasl(self, series):
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'cbf1'), exist_ok=True)
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'm0'), exist_ok=True)
        series.pop(0)
        keys = list(series.keys())
        keys.sort()
        image_data = pydicom.dcmread(series[keys[0]])
        self._get_format_pack(image_data)

        m0 = image_data.pixel_array
        SorterUtil.demosaic(image_data, m0, 'asl-m0', 'm0', self.intermediate_folder, self.formatpack)

        pwi = np.zeros(np.shape(m0))
        repetitions = int((len(keys) - 1) / 2)

        for file_idx in range(1, repetitions * 2, 2):
            label = pydicom.dcmread(series[keys[file_idx]]).pixel_array.astype(float)
            control = pydicom.dcmread(series[keys[file_idx + 1]]).pixel_array.astype(float)
            pwi += (control - label) / repetitions

        # calculating using simple model
        extra_factor = 2  # TODO 需要用更多数据摸索，这是友谊医院数据
        u_cnv = 6000  # units conversion from mL/g/s to mL/100g/min
        bb_p = 0.9  # mL/g, blood-brain partition coefficient
        t_inv = self.delay_time  # sec, inversion time
        t1_b = 1.65  # sec, relaxation of blood at 3.0T
        lab_eff = 0.98  # labeling efficiency for PASL
        l_dur = self.label_dur  # sec, labeling duration

        m0[m0 < 1] = 1
        pwi[pwi < 0] = 0

        upper_param = u_cnv * bb_p * np.exp(t_inv / t1_b)
        lower_param = 2 * lab_eff * l_dur
        cbf_pasl_calc = pwi / m0 * upper_param / lower_param * extra_factor

        mask = self._gen_mask(m0)
        cbf_pasl_calc = np.multiply(cbf_pasl_calc, mask)
        cbf_pasl_calc = gaussian_filter(cbf_pasl_calc, 0.7)
        cbf_pasl_calc = cbf_pasl_calc.astype(np.int16)
        SorterUtil.demosaic_sitk(image_data, cbf_pasl_calc, 'cbf1', self.intermediate_folder, self.formatpack)

    def generate_cbf_3d_pasl(self, series_m0, series_perf, extra_factor):
        if extra_factor is None or extra_factor == 0:
            raise Exception('extra_factor 未设置')

        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'cbf1'), exist_ok=True)
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'm0'), exist_ok=True)
        series_m0.pop(0)
        series_perf.pop(0)
        keys = list(series_m0.keys())
        keys.sort()
        image_data = pydicom.dcmread(series_m0[keys[0]])
        if "MOSAIC" in image_data[0x0008, 0x0008].value:
            self._get_format_pack(image_data)
            # TODO: extra_factor自己猜的，需要更多的验证，这是苏州第一医院的
            extra_factor = 32 / extra_factor
            m0 = image_data.pixel_array * extra_factor
            SorterUtil.demosaic(image_data, m0, 'asl-m0', 'm0', self.intermediate_folder, self.formatpack)
        else:
            for slice_idx in range(len(keys)):
                image_data = pydicom.dcmread(series_m0[keys[slice_idx]])
                DCMUtil.copy_dcm_slice(image_data, self.intermediate_folder, 'asl', 'm0', slice_idx)
                image_arr = image_data.pixel_array
                if slice_idx == 0:
                    [a, b] = np.shape(image_arr)
                    m0 = np.zeros((a, b, len(keys)))
                # TODO: extra_factor自己猜的，需要更多验证，这是一脉阳光的
                extra_factor = 32 / extra_factor
                m0[:, :, slice_idx] = image_arr * extra_factor

            DCMUtil.dcm_to_nii(self.intermediate_folder, 'asl', 'm0')
            m0 = SorterUtil.to_mosaic(m0)

        keys = list(series_perf.keys())
        keys.sort()
        image_data = pydicom.dcmread(series_perf[keys[0]])
        self._get_format_pack(image_data)

        pwi = np.zeros(np.shape(m0))
        repetitions = int(len(keys) / 2)
        for file_idx in range(0, repetitions * 2, 2):
            control = pydicom.dcmread(series_perf[keys[file_idx]]).pixel_array.astype(float)
            label = pydicom.dcmread(series_perf[keys[file_idx + 1]]).pixel_array.astype(float)
            pwi += (control - label) / repetitions

        # calculating using simple model
        u_cnv = 6000  # units conversion from mL/g/s to mL/100g/min
        bb_p = 0.9  # mL/g, blood-brain partition coefficient
        t_inv = self.delay_time  # sec, inversion time
        t1_b = 1.65  # sec, relaxation of blood at 3.0T
        lab_eff = 0.98  # labeling efficiency for PASL
        l_dur = self.label_dur  # sec, labeling duration

        m0[m0 < 1] = 1
        mask = self._gen_mask(m0)

        check = np.multiply(mask, pwi)

        if sum(check.flatten()) < 0:
            pwi = -pwi
        pwi[pwi < 0] = 0

        upper_param = u_cnv * bb_p * np.exp(t_inv / t1_b)
        lower_param = 2 * lab_eff * l_dur
        cbf_pasl_calc = pwi / m0 * upper_param / lower_param
        cbf_pasl_calc = np.multiply(cbf_pasl_calc, mask)
        cbf_pasl_calc = gaussian_filter(cbf_pasl_calc, 0.7)
        cbf_pasl_calc = cbf_pasl_calc.astype(np.int16)
        SorterUtil.demosaic_sitk(image_data, cbf_pasl_calc, 'cbf1', self.intermediate_folder, self.formatpack)

    def parse_pasl_from_archive_folder(self, series):
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'mcbf'), exist_ok=True)
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'att'), exist_ok=True)
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'acbv'), exist_ok=True)
        os.makedirs(os.path.join(self.intermediate_folder, 'asl', 'm0'), exist_ok=True)
        series.pop(0)
        for file_idx in series.keys():
            image_data = pydicom.dcmread(series[file_idx])
            self._get_format_pack(image_data)
            if image_data[0x0020, 0x4000].value == 'M0 Image':
                SorterUtil.demosaic(image_data, image_data.pixel_array, 'asl-m0', 'm0',
                                    self.intermediate_folder,
                                    self.formatpack)
            elif image_data[0x0020, 0x4000].value == 'relCBF':
                cbf_arr = image_data.pixel_array
                SorterUtil.demosaic(image_data, image_data.pixel_array, 'asl-mcbf', 'mcbf',
                                    self.intermediate_folder,
                                    self.formatpack)
            elif image_data[0x0020, 0x4000].value == 'Bolus Arrival Time (BAT) Image':
                att_arr = image_data.pixel_array
                if [0x0028, 0x1053] in image_data:
                    image_data[0x0028, 0x1053].value = 0.001
                else:
                    image_data.add_new([0x0028, 0x1053], 'DS', 0.001)
                SorterUtil.demosaic(image_data, image_data.pixel_array, 'asl-att', 'att',
                                    self.intermediate_folder,
                                    self.formatpack)

        image_data = pydicom.dcmread(series[file_idx])
        image_data.add_new([0x0028, 0x1053], 'DS', 0.001)
        SorterUtil.demosaic(image_data, cbf_arr * att_arr / 60, 'asl-acbv', 'acbv',
                            self.intermediate_folder,
                            self.formatpack)

    def siemens_data_check(self, seriesm0):
        self.reader_ref = sitk.ImageFileReader()
        self.reader_ref.LoadPrivateTagsOn()
        try:
            self.reader_ref.SetFileName(seriesm0[1])
            self.ref_image_data = self.reader_ref.Execute()
        except:
            self.reader_ref.SetFileName([i for i in seriesm0.values()][1])
            self.ref_image_data = self.reader_ref.Execute()

        try:
            self.slice_num = int(float(self.reader_ref.GetMetaData("0019|100a")))
        except:
            raise Exception('slice number not found')

        if float(self.reader_ref.GetMetaData("0020|1041")) >= 0:
            self.slice_order = "descending"
        elif float(self.reader_ref.GetMetaData("0020|1041")) < 0:
            self.slice_order = "ascending"

    def _gen_mask(self, input_image):
        threshold = 0.5 * np.sum(np.multiply(input_image, input_image)) / np.sum(input_image)
        output_image = np.zeros(np.shape(input_image))
        output_image[input_image >= threshold] = 1
        return output_image

    def _get_format_pack(self, image_data):
        self.row = image_data[0x0028, 0x0010].value
        self.col = image_data[0x0028, 0x0011].value
        slice_num = image_data[0x0019, 0x100A].value
        sn = math.ceil(np.sqrt(slice_num))
        dmrow = image_data[0x0028, 0x0010].value // sn
        siemensst = image_data[0x0018, 0x0050].value
        siemenssl = image_data[0x0020, 0x1041].value
        siemensipp = image_data[0x0020, 0x0032].value
        siemensiop = [1, 0, 0, 0, 1, 0]
        self.formatpack = [slice_num, sn, dmrow, siemensst, siemenssl, siemensipp, siemensiop]

    def Sorter(self, intermediate_folder, series_asl, delay_time, delay_rep, label_dur, extra_factor, debugMode=False):
        self.intermediate_folder = intermediate_folder
        self.debugMode = debugMode
        if type(delay_time) == list:
            self.delay_time = delay_time[0]
        elif type(delay_time) == float:
            self.delay_time = delay_time
        if type(label_dur) == list:
            self.label_dur = label_dur[0]
        elif type(label_dur) == float:
            self.label_dur = label_dur

        os.makedirs(os.path.join(self.intermediate_folder, 'asl'), exist_ok=True)
        if "PARAMETER_FIT" in series_asl.keys():
            self.parse_pasl_from_archive_folder(series_asl["PARAMETER_FIT"][0])
        elif "2DPASL_RAW" in series_asl.keys():
            self.generate_cbf_2d_pasl(series_asl["2DPASL_RAW"][0])
        elif "3DPASL_PERF" in series_asl.keys():
            if "3DPASL_M0" in series_asl.keys():
                self.generate_cbf_3d_pasl(series_asl["3DPASL_M0"][0], series_asl["3DPASL_PERF"][0], extra_factor)
            else:
                self.generate_cbf_3d_pasl(series_asl["3DPASL_PERF"][0].copy(), series_asl["3DPASL_PERF"][0], extra_factor)

        # self.seriesdelay = seriesdelay
        # self.delay_rep = delay_rep
        # self.delay_num = len(init_delay)
        # self.init_delay = float(init_delay[0])  # 导进来的init_delay其实是每一个delay的时间点
        # if self.delay_num == 1:
        #     self.interval = float(0)
        # else:
        #     self.interval = float(init_delay[1] - init_delay[0])
        # self.debugMode = debugMode
        # if not os.path.exists(os.path.join(self.intermediate_folder, 'asl')):
        #     os.mkdir(os.path.join(self.intermediate_folder, 'asl'))
        #
        # if self.seriesdelay == 2:
        #     self.parse_pasl_from_archive_folder(seriesm0)
        # elif self.seriesdelay == 1:
        #     self.parse_2Dpasl_from_archive_folder(seriesm0)
        # elif (self.seriesdelay == 5) or (self.seriesdelay == 1):
        #     self.siemens_data_check(seriesm0)
        #     self.generate_from_raw_folder(seriesm0)



import os

import numpy
import numpy as np
import pydicom

from log import logger

from ._dicom import SeriesModel, DicomRequiredTags
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterBase
from .errors import *
from ._type import *
from typing import Union


class DataSorterStruct(DataSorterBase):
    def __init__(self, data_type, output_root, series_list):
        DataSorterBase.__init__(self, data_type, output_root, series_list)

    @staticmethod
    def validate_series(series_list: List[SeriesModel]):
        # 多个struct，优先选择轴状位的
        struct: SeriesModel
        # 只有一个轴位的，没得选
        if len(series_list) == 1:
            struct = series_list[0]
        else:
            # 多个选择其一使用
            struct_tra_list = []
            for s in series_list:
                if s.SliceDirection == SliceDirection.Axial:
                    struct_tra_list.append(s)
            if len(struct_tra_list) == 1:
                struct = struct_tra_list[0]
            elif len(struct_tra_list) > 1:
                # 优先选择分辨率高的、SeriesNumber号小的（号大的有可能是计算而来的）
                struct = sorted(struct_tra_list, key=lambda x: (x.SpacingBetweenSlices, x.SeriesNumber))[0]
            else:
                # 没有轴状位的序列
                struct = series_list[0]
        return struct

'''
这里是除了ASL以外的MR图像
T1、T2、T2 FLAIR、DWI
'''
class DataSorterT1(DataSorterStruct):
    def __init__(self, data_type, output_root, series_list):
        DataSorterBase.__init__(self, data_type, output_root, series_list)

    def Sorter(self):
        super(DataSorterT1, self).Sorter()

        self.T1 = self.validate_series(self.series_list)
        logger.debug(f"清洗的T1：{self.T1.SeriesDescription}，剖面：{self.T1.SliceDirection}，"
                         f"分辨率：{self.T1.PixelSpacing[0]}*{self.T1.PixelSpacing[1]}*{self.T1.SpacingBetweenSlices}")

        cur_series_number = SeriesNumberStartWith.T1
        T1_img = self.T1.load()[...,0]
        write_dicom_series(T1_img, self.T1, 't1', cur_series_number, os.path.join(self.output_root, 'struct', f't1'))

class DataSorterT2(DataSorterStruct):
    def __init__(self, data_type, output_root, series_list:List[SeriesModel]):
        DataSorterBase.__init__(self, data_type, output_root, series_list)

    def Sorter(self):
        super(DataSorterT2, self).Sorter()

        self.T2 = self.validate_series(self.series_list)
        logger.debug(f"清洗的T2：{self.T2.SeriesDescription}，剖面：{self.T2.SliceDirection}，"
                         f"分辨率：{self.T2.PixelSpacing[0]}*{self.T2.PixelSpacing[1]}*{self.T2.SpacingBetweenSlices}")

        cur_series_number = SeriesNumberStartWith.T2
        T2_img = self.T2.load()[...,0]
        write_dicom_series(T2_img, self.T2, 't2', cur_series_number, os.path.join(self.output_root, 'struct', f't2'))


class DataSorterT2FLAIR(DataSorterStruct):
    def __init__(self, data_type, output_root, series_list):
        DataSorterBase.__init__(self, data_type, output_root, series_list)

    def Sorter(self):
        super(DataSorterT2FLAIR, self).Sorter()

        self.T2FLAIR = self.validate_series(self.series_list)
        logger.debug(f"清洗的T2FLAIR：{self.T2FLAIR.SeriesDescription}，剖面：{self.T2FLAIR.SliceDirection}，"
                         f"分辨率：{self.T2FLAIR.PixelSpacing[0]}*{self.T2FLAIR.PixelSpacing[1]}*{self.T2FLAIR.SpacingBetweenSlices}")

        cur_series_number = SeriesNumberStartWith.T2FLAIR
        T2FLAIR_img = self.T2FLAIR.load()[..., 0]
        write_dicom_series(T2FLAIR_img, self.T2FLAIR, 't2 flair', cur_series_number, os.path.join(self.output_root, 'struct', f't2 flair'))

class DataSorterDWI(DataSorterBase):
    def __init__(self, data_type, output_root, series_list: List[SeriesModel]):
        DataSorterBase.__init__(self, data_type, output_root, series_list)

        self.DWI_list:List[SeriesModel] = []
        self.ADC_list:List[SeriesModel] = []
        reg_DWI_list:List[SeriesModel] = []
        for series in self.series_list:
            series_description:str = series.SeriesDescription.lower()
            if series_description.find('adc') > -1 or series_description.find('apparent diffusion coefficient') > -1:
                if series_description.find('eadc') > -1 or series_description.find('exponential') > -1:
                    continue
                self.ADC_list.append(series)
            elif series_description.find('dwi') > -1 or series_description.find('tracew') > -1:
                if series_description.find('reg') > -1:
                    reg_DWI_list.append(series)
                else:
                    self.DWI_list.append(series)
        if len(self.DWI_list) == 0:
            self.DWI_list = reg_DWI_list
        self.series_list = self.DWI_list + self.ADC_list

    def validate_series(self):
        if len(self.DWI_list) < 1:
            raise MissingSequenceError('DWI')
        elif len(self.DWI_list) > 1:
            logger.warning(f'发现{len(self.DWI_list)}组DWI数据')
        self.DWI = self.DWI_list[0]

        if len(self.ADC_list) < 1:
            logger.warning(f'未发现ADC序列')
            self.ADC: SeriesModel = None
            # raise MissingSequenceError('ADC')
        else:
            logger.warning(f'发现{len(self.DWI_list)}组ADC数据')
            self.ADC = self.ADC_list[0]

        if self.ADC is not None and self.DWI.SliceNumber != self.ADC.SliceNumber:
            raise SpaceIsDifferentError('DWI', 'ADC')
        if self.ADC is not None and (self.DWI.Repetition != 2 or self.ADC.Repetition != 1):
            raise SliceLessError(self.ADC.SliceNumber)

    def Sorter(self):
        super(DataSorterDWI, self).Sorter()
        self.validate_series()

        # DWI需要包含B0与B1000
        # 标准dicom中的B值(弥散系数)标签信息
        second_slice_number = 1 if self.DWI.is_second_arrangement_mode else self.DWI.SliceNumber

        def get_tag(dcm_obj: Union[DicomRequiredTags, pydicom.Dataset], key, default=None):
            if isinstance(dcm_obj, DicomRequiredTags):
                return dcm_obj.get_tag(key, default=default)
            else:
                return DicomRequiredTags.get_key(key, dcm_obj, default)

        # 3D dicom的B值在PerFrameFunctionalGroupsSequence中
        if self.DWI.dicom_image_type == DICOM_IMAGE_TYPE.DICOM_IMAGE_TYPE_3D:
            first_instance = get_tag(self.DWI.PerFrameFunctionalGroupsSequence[0], (0x0018, 0x9117))[0]
            second_instance = get_tag(self.DWI.PerFrameFunctionalGroupsSequence[second_slice_number], (0x0018, 0x9117))[0]
        else:
            first_instance = self.DWI[0]
            second_instance = self.DWI[second_slice_number]

        first_b_value = get_tag(first_instance, (0x0018, 0x9087), 0)
        second_b_value = get_tag(second_instance, (0x0018, 0x9087), 0)
        if self.DWI.Manufacturer == MANUFACTURER.kMANUFACTURER_GE:
            first_b_value = first_b_value if first_b_value != 0 else \
                get_tag(first_instance, (0x0043, 0x1039), [0])[0]
            second_b_value = second_b_value if second_b_value != 0 else \
                get_tag(second_instance, (0x0043, 0x1039), [0])[0]

        elif self.DWI.Manufacturer == MANUFACTURER.kMANUFACTURER_SIEMENS:
            first_b_value = first_b_value if first_b_value != 0 else \
                get_tag(first_instance, (0x0019, 0x100C), 0)
            second_b_value = second_b_value if second_b_value != 0 else \
                get_tag(second_instance, (0x0019, 0x100C), 0)

        elif self.DWI.Manufacturer == MANUFACTURER.kMANUFACTURER_PHILIPS:
            pass
            # first_b_value = self.DWI[0].get_tag((0x0018, 0x9087), 0)
            # second_b_value = self.DWI[self.DWI.SliceNumber].get_tag((0x0018, 0x9087), 0)

        else:
            raise UnSupportDataTypeError(self.DWI.Manufacturer)

        first_b_value = int(first_b_value)
        second_b_value = int(second_b_value)
        if first_b_value == 0 and second_b_value == 0:
            raise DiffusionBValueMissingError()

        DWI_img = self.DWI.load()
        cur_series_number = SeriesNumberStartWith.DWI

        write_dicom_series(DWI_img[...,0], self.DWI, f'dwi-b{first_b_value}', cur_series_number,
                           os.path.join(self.output_root, 'dwi', f'b{first_b_value}'))
        cur_series_number += 1
        write_dicom_series(DWI_img[...,1], self.DWI, f'dwi-b{second_b_value}', cur_series_number,
                               os.path.join(self.output_root, 'dwi', f'b{second_b_value}'))

        cur_series_number += 1
        rescale_type = '10^-6 mm^2/s'
        if self.ADC is not None:
            # ADC， 确定单位
            rescale_type_1 = self.ADC.get_tag((0x0028, 0x1054), '')
            rescale_type_2 = self.ADC.get_tag((0x2005, 0x140B), '')
            if rescale_type_1 != '':
                rescale_type = rescale_type_1
            elif rescale_type_2 != '':
                rescale_type = rescale_type_2

            rescale = 1000 if rescale_type == '10^-3 mm^2/s' else 1
            ADC_img = self.ADC.load(rescale)[...,0]
            write_dicom_series(ADC_img, self.DWI, 'dwi-adc', cur_series_number,
                               os.path.join(self.output_root, 'dwi', 'adc'))

        else:
            logger.debug(f'开始计算ADC')
            # 使second_b_value等于较大B值的那个图像
            first_b0_idx, second_b1_idx = 0, 1
            if first_b_value > second_b_value:
                first_b_value, second_b_value = second_b_value, first_b_value
                first_b0_idx, second_b1_idx = second_b1_idx, first_b0_idx
            ADC_img = np.log(DWI_img[...,first_b0_idx] / DWI_img[...,second_b1_idx]) / (second_b_value - first_b_value)
            ADC_img[np.isinf(ADC_img) | np.isnan(ADC_img)] = 0
            ADC_img *= 10**6
            ADC_img[ADC_img < 0] = -256

            write_dicom_series(ADC_img, self.DWI, 'dwi-adc', cur_series_number,
                               os.path.join(self.output_root, 'dwi', 'adc'))


from ._dicom import *
from ._dicom_util import write_dicom_series
from .data_sorter_base import DataSorterASLBase


class DataSorterUIHPASL(DataSorterASLBase):
    def __init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, extra_factor, calc_c_cbf):
        DataSorterASLBase.__init__(self, data_type, output_root, delay_time, delay_rep, label_dur, series_list, calc_c_cbf)
        if extra_factor is None or extra_factor == 0:
            logger.debug('extra_factor 未设置')
            # raise Exception('extra_factor 未设置')
        else:
            self.extra_factor = 32 / extra_factor
        self.M0 = None
        self.CBF = None # CBF参数图
        self.Control = None
        self.Label = None

        self.M0_count = 0
        self.CBF_count = 0
        # 区别M0与灌注图
        for series in self.series_list:
            if not isinstance(series, SeriesModel):
                continue
            seriesDescription = series.SeriesDescription.lower().replace('-', ' ')
            if seriesDescription.find('m0') > -1:
                self.M0_count += 1
                self.M0 = series
            elif seriesDescription.find('cbf') > -1:
                self.CBF_count += 1
                self.CBF = series
            elif seriesDescription.find('ctrl') > -1:
                self.Control = series
            elif seriesDescription.find('tag') > -1:
                self.Label = series

    def validate_series(self):
        if self.M0 is None:
            raise MissingSequenceError('M0')
        if self.CBF is None:
            raise MissingSequenceError('CBF')

        if self.M0_count > 1:
            raise SeriesTooMany(f'发现{self.M0_count}组M0图像，请保留一组再进行处理')
        if self.CBF_count > 1:
            raise SeriesTooMany(f'发现{self.CBF_count}组CBF图像，请保留一组再进行处理')

    def Sorter(self):
        super(DataSorterUIHPASL, self).Sorter()

        self.validate_series()

        # self.generate_calc_cbf_3d_pasl()
        self.generate_cbf_3d_pasl()

    def generate_cbf_3d_pasl(self):
        self.M0: SeriesModel
        m0_image = self.M0.load()
        # 多张M0只有第一张有效
        m0_image = m0_image[...,0]
        write_dicom_series(m0_image, self.M0, 'asl-m0', 1001, os.path.join(self.output_root, 'asl', 'm0'))

        self.CBF: SeriesModel
        cbf_image = self.CBF.load()
        cbf_image = cbf_image[...,0]
        # todo M0与CBF分辨率不一致时
        # if cbf_image.shape != m0_image.shape:
        #     raise SpaceIsDifferentError('CBF', 'M0')
        write_dicom_series(cbf_image * 10, self.CBF, 'asl-cbf1', 1002, os.path.join(self.output_root, 'asl', 'cbf1'),
                           slope=0.1)

    def generate_calc_cbf_3d_pasl(self):
        self.M0: SeriesModel
        m0_image = self.M0.load()
        # 多张M0只有第一张有效
        m0_image = m0_image[:,:,:,0]
        write_dicom_series(m0_image, self.M0, 'asl-m0', 1001, os.path.join(self.output_root, 'asl', 'm0'))

        control = self.Control.load()
        label = self.Label.load()
        repetitions = control.shape[-1]
        pwi = np.zeros(m0_image.shape)
        sub = control - label
        for i in range(repetitions):
            pwi += sub[:,:,:,i]
        pwi /= repetitions

        # calculate CBF
        t_inv = self.delay_time  # sec, inversion time
        l_dur = self.label_dur  # sec, labeling duration
        bb_p = 0.9  # mL/g, blood-brain partition coefficient
        t1_b = 1.65  # sec, relaxation of blood at 3.0T
        lab_eff = 0.98  # labeling efficiency for PASL
        u_cnv = 6000  # units conversion from mL/g/s to mL/100g/min

        upper_param = u_cnv * bb_p * np.exp(t_inv / t1_b)
        lower_param = 2 * lab_eff * l_dur
        calc_cbf = pwi / m0_image.astype(float) * upper_param / lower_param
        write_dicom_series(calc_cbf * 10, self.M0, 'asl-cbf1', 1002, os.path.join(self.output_root, 'asl', 'cbf1'),
                           slope=0.1)

    def _gen_mask(self, input_image):
        threshold = 0.5 * np.sum(np.multiply(input_image, input_image)) / np.sum(input_image)
        output_image = np.zeros(np.shape(input_image))
        output_image[input_image >= threshold] = 1
        return output_image